import torch
import numpy as np
import torch.nn.functional as F
import pdb
from instanseg.utils.utils import show_images, timer
from tqdm import tqdm

def edge_mask(labels, ignore=[None]):
    labels = labels.squeeze()
    first_row = labels[0, :]
    last_row = labels[-1, :]
    first_column = labels[:, 0]
    last_column = labels[:, -1]

    edges = []
    if 'top' not in ignore:
        edges.append(first_row)
    if 'bottom' not in ignore:
        edges.append(last_row)
    if 'left' not in ignore:
        edges.append(first_column)
    if 'right' not in ignore:
        edges.append(last_column)

    edges = torch.cat(edges, dim=0)
    return torch.isin(labels, edges[edges > 0])


def remove_edge_labels(labels, ignore=[None]):
    return labels * ~edge_mask(labels, ignore=ignore)


def _to_shape(a, shape):
    """Pad a tensor to a given shape."""
    if len(a.shape) == 2:
        a = a.unsqueeze(0)
    y_, x_ = shape
    y, x = a[0].shape[-2:]
    y_pad = max(0, y_ - y)
    x_pad = max(0, x_ - x)
    return torch.nn.functional.pad(a, (x_pad // 2, x_pad // 2 + x_pad % 2, y_pad // 2, y_pad // 2 + y_pad % 2))


def _to_shape_bottom_left(a, shape):
    """Pad a tensor to a given shape."""
    if len(a.shape) == 2:
        a = a.unsqueeze(0)
    y_, x_ = shape
    y, x = a[0].shape[-2:]
    y_pad = max(0, y_ - y)
    x_pad = max(0, x_ - x)
    return torch.nn.functional.pad(a, (0, x_pad, 0, y_pad))

def chops(img_shape: tuple, shape: tuple, overlap: int = 0) -> tuple:
    """This function splits an image into desired windows and returns the indices of the windows"""

    if (torch.tensor(img_shape[-2:]) < torch.tensor(shape)).any():
        return [0], [0]
    h, v = img_shape[-2:]


    stride_h = shape[0] - overlap
    stride_v = shape[1] - overlap

        
    max_v = (v - shape[1])
    max_h = (h - shape[0])

    v_index = np.array([i * stride_v for i in range(v // stride_v )] + [v - shape[1]])
    h_index = np.array([i * stride_h for i in range(h // stride_h )] + [h - shape[0]])

    v_index = v_index[v_index <= max_v]
    h_index = h_index[h_index <= max_h]

    v_index = np.unique(v_index)
    h_index = np.unique(h_index)

    return h_index, v_index


def tiles_from_chops(image: torch.Tensor, shape: tuple, tuple_index: tuple) -> list:
    """This function takes an image, a shape, and a tuple of window indices (e.g., output of the function chops)
    and returns a list of windows"""
    h_index, v_index = tuple_index

    if len(image.shape) == 2:
        image = image.unsqueeze(0)

    stride_h = shape[0]
    stride_v = shape[1]
    tile_list = []
    for i, window_i in enumerate(h_index):
        for j, window_j in enumerate(v_index):
            current_window = image[..., window_i:window_i + stride_h, window_j:window_j + stride_v]
            tile_list.append(current_window)
    return tile_list


def stitch(tiles: list, shape: tuple, chop_list: list, final_shape: tuple, offset : int):
    """This function takes a list of tiles, a shape, and a tuple of window indices (e.g., outputed by the function chops)
    and returns a stitched image"""
    from instanseg.utils.pytorch_utils import torch_fastremap, match_labels

    canvas = torch.zeros(final_shape, dtype=torch.int32, device = tiles[0].device)

    running_max = 0
            
    for i, window_i in enumerate(chop_list[0]):
        for j, window_j in enumerate(chop_list[1]):

            edge_window = False
            new_tile = tiles[i * len(chop_list[1]) + j][None]
            ignore_list = []
            if i == 0:
                ignore_list.append("top")
            if j == 0:
                ignore_list.append("left")
            if i == len(chop_list[0])-1:
                ignore_list.append("bottom")
            if j == len(chop_list[1])-1:
                ignore_list.append("right")

            if i == len(chop_list[0])-1 and j == len(chop_list[1])-1:
                edge_window = True
                tile1 = canvas[..., window_i + offset:window_i + shape[0], window_j + offset:window_j + shape[1]]
                tile2 = remove_edge_labels(new_tile[:,offset:shape[0],offset: shape[1]], ignore = ignore_list)

            elif i == len(chop_list[0])-1:
                edge_window = True
                tile1 = canvas[..., window_i + offset:window_i + shape[0], window_j + offset:window_j + shape[1]]
                tile2 = remove_edge_labels(new_tile[:,offset:shape[0],offset : shape[1]], ignore =ignore_list)

            elif j == len(chop_list[1])-1:
                edge_window = True
                tile1 = canvas[..., window_i + offset:window_i + shape[0], window_j + offset:window_j + shape[1]]
                tile2 = remove_edge_labels(new_tile[:,offset:shape[0],offset: shape[1]], ignore = ignore_list)

            if i == 0 and j == 0:
                edge_window = True
                tile1 = canvas[..., window_i  :window_i + shape[0], window_j :window_j + shape[1]]
                tile2 = remove_edge_labels(new_tile[:, :shape[0], : shape[1]], ignore = ignore_list)
            elif i == 0:
                edge_window = True
                tile1 = canvas[..., window_i  :window_i + shape[0], window_j + offset :window_j + shape[1]]
                tile2 = remove_edge_labels(new_tile[:, :shape[0],offset : shape[1]], ignore = ignore_list)

            elif j == 0:
                edge_window = True
                tile1 = canvas[..., window_i  + offset:window_i + shape[0], window_j:window_j + shape[1]]
                tile2 = remove_edge_labels(new_tile[:,offset :shape[0],: shape[1]], ignore = ignore_list)

            if edge_window:

                tile2 = torch_fastremap(tile2)
                tile2[tile2>0] = tile2[tile2>0] + running_max

                remapped = match_labels(tile1, tile2, threshold = 0.1)[1]
                tile1[remapped>0] = remapped[remapped>0].int()

                running_max = max(running_max, tile1.max())

            else:
                
                tile1 = canvas[..., window_i + offset:window_i + shape[0] - offset, window_j + offset:window_j + shape[1] - offset]
                tile2 = remove_edge_labels(new_tile[:,offset:shape[0] -offset,offset: shape[1]-offset])


                tile2 = torch_fastremap(tile2)
                tile2[tile2>0] = tile2[tile2>0] + running_max

                remapped = match_labels(tile1, tile2, threshold = 0.1)[1]
                tile1[remapped>0] = remapped[remapped>0].int()

                running_max = max(running_max, tile1.max())
            

    return canvas


def zarr_to_json_export(path_to_zarr, cell_size = 30, size = 1024, scale = 1, n_dim = 1):

    import zarr
    import numpy as np
    import os

    z = zarr.open(path_to_zarr, mode='r')

    output_path = str(path_to_zarr).replace(".zarr",".geojson")   


    from instanseg.utils.tiling import chops

    chop_list = chops(z.shape, shape=(size,size), overlap=cell_size)

    from instanseg.utils.utils import labels_to_features
    from instanseg.utils.tiling import remove_edge_labels
    import json

    count = 0
    #Delete the file if it already exists
    if os.path.exists( output_path):
        os.remove( output_path)

    with open(os.path.join(output_path), "a") as outfile:
        outfile.write('[')
        outfile.write('\n')

    if n_dim == 1:
        classes = [None]
    else:
        classes = ["Nucleus","Cell"]


    for i, window_i in tqdm(enumerate(chop_list[0]), total = len(chop_list[0])):
        for j, window_j in enumerate(chop_list[1]):
            for n in range(n_dim):
                
                image = z[n,window_i:window_i+size, window_j:window_j+size]
              
                image = remove_edge_labels(torch.tensor(image)).numpy()

                features = labels_to_features(image.astype(np.int32), object_type='detection', include_labels=True,
                                        classification=classes[n],offset=[window_j*scale,window_i*scale], downsample = scale)


            if features != []:
                count+=1
                geojson = json.dumps(features)

                with open(os.path.join( output_path), "a") as outfile:
                    outfile.write(geojson[1:-1] + ",")
                    outfile.write('\n')

    with open(os.path.join( output_path), "a") as outfile:
        outfile.write(']')


def segment_image_larger_than_memory(instanseg_class, # instanseg class method
                                    image_path: str,  
                                     memory_block_size: tuple = (3000,3000), #this is the size of the image that will be read in memory
                                     inference_tile_size: tuple = (512,512), #this is the size of the image that will be passed to the model
                                     overlap = 100,
                                    cell_size = 20, 
                                    to_geojson = False, 
                                    batch_size = 3,
                                    prediction_tag: str = "_instanseg_prediction",
                                    pixel_size: float = None,
                                    normalisation_subsampling_factor = 10,
                                    **kwargs):
    
    """This function uses slideio to read an image and then segments it using the instanseg model. 
    The segmentation is done in a tiled manner to avoid memory issues. 
    The function returns a zarr file with the segmentation. The zarr file is saved in the same directory as the image with the same name but with the extension .zarr. 
    The function also returns the zarr file object."""

    import zarr
    from itertools import product
    from instanseg.utils.pytorch_utils import torch_fastremap, match_labels
    from pathlib import Path
    
    instanseg = instanseg_class.instanseg

    image_path, img_pixel_size = instanseg_class.read_image(image_path)
    slide = instanseg_class.read_slide(image_path)

    n_dim = 2 if instanseg.cells_and_nuclei else 1
    model_pixel_size = instanseg.pixel_size

    new_stem = Path(image_path).stem + prediction_tag
    file_with_zarr_extension = Path(image_path).parent / (new_stem + ".zarr")


    if img_pixel_size > 1 or img_pixel_size < 0.1:
        import warnings
        warnings.warn("The image pixel size {} is not in microns.".format(img_pixel_size))
        if pixel_size is not None:
            img_pixel_size = pixel_size
        else:
            raise ValueError("The image pixel size {} is not in microns.".format(img_pixel_size))
    
    scale_factor = model_pixel_size/img_pixel_size

    dims = slide.dimensions
    dims = (int(dims[1]/ scale_factor), int(dims[0]/scale_factor)) #The dimensions are opposite to numpy/torch/zarr dimensions.

    pad2 = overlap + cell_size
    pad = overlap

    shape = memory_block_size
    
    chop_list = chops(dims, shape, overlap=2*pad2)

    chunk_shape = (n_dim,shape[0],shape[1])
    store = zarr.DirectoryStore(file_with_zarr_extension) 
    canvas = zarr.zeros((n_dim,dims[0],dims[1]), chunks=chunk_shape, dtype=np.int32, store=store, overwrite = True)

    running_max = 0

    total = len(chop_list[0]) * len(chop_list[1])
    for _, ((i, window_i), (j, window_j)) in tqdm(enumerate(product(enumerate(chop_list[0]), enumerate(chop_list[1]))), total=total, colour = "green", desc = "Slide progress: "):

      #  input_data = scene.read_block((int(window_j*scale_factor), int(window_i*scale_factor), int(shape[0]*scale_factor), int(shape[1]*scale_factor)), size = shape)

        best_level = slide.get_best_level_for_downsample(scale_factor)
        downsample_factor = slide.level_downsamples[best_level]

        initial_pixel_size = img_pixel_size
        itermediate_pixel_size = initial_pixel_size * downsample_factor
        final_pixel_size = model_pixel_size

        intermediate_to_final = final_pixel_size/itermediate_pixel_size
        
        # Calculate the size of the region needed at the base level to get the desired output size
        intermediate_shape = (int(shape[0] * intermediate_to_final), int(shape[1] * intermediate_to_final))

        input_data = slide.read_region((int(window_j*scale_factor), int(window_i*scale_factor)), best_level, (int(intermediate_shape[0]) , int(intermediate_shape[1])), as_array=True)
    
        input_tensor = instanseg_class._to_tensor(input_data)

        new_tile = instanseg_class.eval_medium_image(input_tensor,
                                          pixel_size = itermediate_pixel_size,
                                          tile_size = inference_tile_size[0],
                                          batch_size = batch_size,
                                          return_image_tensor = False,
                                          normalisation_subsampling_factor = normalisation_subsampling_factor,
                                          )
        
    

        from torch.nn.functional import interpolate
        new_tile = interpolate(new_tile, size=shape[-2:], mode="nearest").int()[0]

       # pdb.set_trace()

        num_iter = new_tile.shape[0]

        for n in range(num_iter):
            
            ignore_list = []
            if i == 0:
                ignore_list.append("top")
            if j == 0:
                ignore_list.append("left")
            if i == len(chop_list[0])-1:
                ignore_list.append("bottom")
            if j == len(chop_list[1])-1:
                ignore_list.append("right")

            if i == len(chop_list[0])-1 and j == len(chop_list[1])-1:
                tile1 = canvas[n, window_i + pad:window_i + shape[0], window_j + pad:window_j + shape[1]]
                tile2 = remove_edge_labels(new_tile[n,pad:shape[0],pad: shape[1]], ignore = ignore_list)

            elif i == len(chop_list[0])-1:
                tile1 = canvas[n, window_i + pad:window_i + shape[0], window_j + pad:window_j + shape[1]]
                tile2 = remove_edge_labels(new_tile[n,pad:shape[0],pad : shape[1]], ignore =ignore_list)

            elif j == len(chop_list[1])-1:
                tile1 = canvas[n, window_i + pad:window_i + shape[0], window_j + pad:window_j + shape[1]]
                tile2 = remove_edge_labels(new_tile[n,pad:shape[0],pad: shape[1]], ignore = ignore_list)

            elif i == 0 and j == 0:
                tile1 = canvas[n, window_i  :window_i + shape[0], window_j :window_j + shape[1]]
                tile2 = remove_edge_labels(new_tile[n, :shape[0], : shape[1]], ignore = ignore_list)
            elif i == 0:
                tile1 = canvas[n, window_i  :window_i + shape[0], window_j + pad :window_j + shape[1]]
                tile2 = remove_edge_labels(new_tile[n, :shape[0],pad : shape[1]], ignore = ignore_list)

            elif j == 0:
                tile1 = canvas[n, window_i  + pad:window_i + shape[0], window_j:window_j + shape[1]]
                tile2 = remove_edge_labels(new_tile[n,pad :shape[0],: shape[1]], ignore = ignore_list)

            if j == 0 or i == 0 or j == len(chop_list[1])-1 or i == len(chop_list[0])-1:

                tile2 = torch_fastremap(tile2)
                tile2[tile2>0] = tile2[tile2>0] + running_max

                tile1_torch = torch.tensor(np.array(tile1), dtype = torch.int32)

                remapped = match_labels(tile1_torch, tile2, threshold = 0.1)[1]
                tile1_torch[remapped>0] = remapped[remapped>0].int()

                running_max = max(running_max, tile1_torch.max())

                if i == len(chop_list[0])-1 and j == len(chop_list[1])-1:
                    canvas[n, window_i + pad:window_i + shape[0], window_j + pad:window_j + shape[1]] = tile1_torch.numpy().astype(np.int32)
                elif i == len(chop_list[0])-1:
                    canvas[n, window_i + pad:window_i + shape[0], window_j + pad:window_j + shape[1]] = tile1_torch.numpy().astype(np.int32)
                elif j == len(chop_list[1])-1:
                    canvas[n, window_i + pad:window_i + shape[0], window_j + pad:window_j + shape[1]] = tile1_torch.numpy().astype(np.int32)
                elif i == 0 and j == 0:
                    canvas[n, window_i  :window_i + shape[0], window_j :window_j + shape[1]] = tile1_torch.numpy().astype(np.int32)
                elif i == 0:
                    canvas[n, window_i  :window_i + shape[0], window_j + pad :window_j + shape[1]] = tile1_torch.numpy().astype(np.int32)
                elif j == 0:
                    canvas[n, window_i  + pad:window_i + shape[0], window_j:window_j + shape[1]] = tile1_torch.numpy().astype(np.int32)

            else:
                
                tile1 = canvas[n, window_i + pad:window_i + shape[0] - pad, window_j + pad:window_j + shape[1] - pad]
                tile2 = remove_edge_labels(new_tile[n,pad:shape[0] -pad,pad: shape[1]-pad])
                
                tile2 = torch_fastremap(tile2)

                tile2[tile2>0] = tile2[tile2>0] + running_max

                tile1_torch = torch.tensor(np.array(tile1), dtype = torch.int32)
                remapped = match_labels(tile1_torch, tile2, threshold = 0.1)[1]

                tile1_torch[remapped>0] = remapped[remapped>0].int()
                running_max = max(running_max, tile1_torch.max())
            
                canvas[n, window_i + pad:window_i + shape[0] - pad, window_j + pad:window_j + shape[1] - pad] = tile1_torch.numpy().astype(np.int32)

    if to_geojson:
        print("Exporting to geojson")
        zarr_to_json_export(file_with_zarr_extension, cell_size = cell_size, size = shape[0], scale = scale_factor, n_dim = n_dim)
            

    return canvas


def instanseg_padding(img: torch.Tensor, extra_pad: int = 0, min_dim: int = 16, ensure_square: bool = False):

    is_square = img.shape[-2] == img.shape[-1]
    original_shape = img.shape[-2:]
    bigger_dim = max(img.shape[-2], img.shape[-1])

    if ensure_square and not is_square:
        img = torch.functional.F.pad(img, [0, bigger_dim - img.shape[-1], 0, bigger_dim - img.shape[-2]], mode='constant')

    padx = min_dim * torch.ceil(torch.tensor((img.shape[-2] / min_dim))).int() - img.shape[-2] + extra_pad * 2
    pady = min_dim * torch.ceil(torch.tensor((img.shape[-1] / min_dim))).int() - img.shape[-1] + extra_pad * 2

    if padx > img.shape[-2]:
        padx = padx - extra_pad
    if pady > img.shape[-1]:
        pady = pady - extra_pad
    img = torch.functional.F.pad(img, [0, int(pady), 0, int(padx)], mode='reflect')

    if ensure_square and not is_square:
        pady = pady + bigger_dim - original_shape[-1]
        padx = padx + bigger_dim - img.shape[-2]
        print(padx, pady)
    

    return img, torch.stack((padx, pady))


def recover_padding(x: torch.Tensor, pad: torch.Tensor):
    # x must be 1,C,H,W or C,H,W
    squeeze = False
    if x.ndim == 3:
        x = x.unsqueeze(0)
        squeeze = True

    if pad[0] == 0:
        pad[0] = -x.shape[2]
    if pad[1] == 0:
        pad[1] = -x.shape[3]

    if squeeze:
        return x[:, :, :-pad[0], :-pad[1]].squeeze(0)
    else:
        return x[:, :, :-pad[0], :-pad[1]]



def sliding_window_inference(input_tensor, predictor, window_size=(512, 512), overlap = 100, max_cell_size = 20, sw_device='cuda',
                             device='cpu', output_channels=1, show_progress = True, batch_size = 1,**kwargs):
    input_tensor = input_tensor.to(device)
    predictor = predictor.to(sw_device)
 
    tuple_index = chops(input_tensor.shape, shape=window_size, overlap=2 * (overlap + max_cell_size))
    tile_list = tiles_from_chops(input_tensor, shape=window_size, tuple_index=tuple_index)
 
 
    #print("Number of tiles: ", len(tile_list))
   # print("Shape of tiles: ", tile_list[0].shape)
 
 
    with torch.no_grad():
        with torch.amp.autocast("cuda"):
            batch_list = [torch.stack(tile_list[batch_size * i:batch_size * (i+1)]) for i in range(int(np.ceil(len(tile_list)/batch_size)))]
            label_list = torch.cat([predictor(tile.to(sw_device),**kwargs).to(device) for tile in tqdm(batch_list, disable= not show_progress,leave = False, colour = "blue")])
 
 
    lab = torch.cat([stitch([lab[i] for lab in label_list],
                            shape=window_size,
                            chop_list=tuple_index,
                            offset = overlap,
                            final_shape=(1, input_tensor.shape[1], input_tensor.shape[2])) 
 
                    for i in range(output_channels)], dim=0)
 
 
    return lab[None]  # 1,C,H,W



import os
import pandas as pd
from tqdm.auto import tqdm
import torch
from pathlib import Path
import argparse
import pdb

parser = argparse.ArgumentParser()
parser.add_argument("-i_p", "--image_path", type=str, default=r"../examples")
parser.add_argument("-m_f", "--model_folder", type=str)
parser.add_argument("-d", "--device", type=str, default=torch.device("cuda:0" if torch.cuda.is_available() else "cpu"))
parser.add_argument("-exclude", "--exclude_str", type=str, default= ["mask","prediction", "geojson", "zip"], help="Exclude files with this string in their name")
parser.add_argument("-pixel_size", "--pixel_size", type=float, default= None, help="Pixel size of the input image in microns")
parser.add_argument("-recursive", "--recursive",default=False, type=lambda x: (str(x).lower() == 'true'),help="Look for images recursively at the image path")
parser.add_argument("-ignore_segmented", "--ignore_segmented",default=False, type=lambda x: (str(x).lower() == 'true'),help="Whether to ignore previously segmented images in the image path")

#advanced usage
parser.add_argument("-tile_size", "--tile_size", type=int, default= 512, help="tile size in pixels given to the model, only used for large images.")
parser.add_argument("-batch_size", "--batch_size", type=int, default= 3, help="batch size, only useful for large images")
parser.add_argument("-save_geojson", "--save_geojson", type=lambda x: (str(x).lower() == 'true'), default= False, help="Output geojson files of the segmentation")

def file_matches_requirement(root,file, exclude_str):
    if not os.path.isfile(os.path.join(root,file)):
        return False
    for e_str in exclude_str:
        if e_str in file:
            return False
        if parser.ignore_segmented:
            if os.path.isfile(os.path.join(root,str(Path(file).stem) + prediction_tag + ".tiff")):
                return False
    return True

prediction_tag = "_instanseg_prediction"


if __name__ == "__main__":
    from instanseg.utils.utils import show_images
    from instanseg.instanseg import InstanSeg

    parser = parser.parse_args()

    if parser.image_path is None or not os.path.exists(parser.image_path):
        from instanseg.utils.utils import drag_and_drop_file
        parser.image_path = drag_and_drop_file()
        print("Using image path: ", parser.image_path)

    if parser.model_folder is None:
        raise ValueError("Please provide a model name")
    
    instanseg = InstanSeg(model_type=parser.model_folder, device= parser.device)
    instanseg.prediction_tag = prediction_tag

    if not parser.recursive:
        print("Loading files from: ", parser.image_path)
        files = os.listdir(parser.image_path)
        files = [os.path.join(parser.image_path, file) for file in files if file_matches_requirement(parser.image_path, file, parser.exclude_str)]
    else:
        print("Loading files recursively from: ", parser.image_path)
        files = []
        for root, dirs, filenames in os.walk(parser.image_path):
            for filename in filenames:
                if file_matches_requirement(root , filename, parser.exclude_str):
                    files.append(os.path.join(root, filename))

    assert len(files) > 0, "No files found in the specified directory"

    
    for file in tqdm(files):

        print("Processing: ", file)

        _ = instanseg.eval(image=file,
                        pixel_size = parser.pixel_size,
                        save_output = True,
                        save_overlay = True,
                        save_geojson = parser.save_geojson,
                        batch_size = parser.batch_size,
                        tile_size = parser.tile_size,)















import os
#from PIL import Image
import tifffile

# the imports for bioimage.io model export
import bioimageio.core
import numpy as np
import torch
import monai
from aicsimageio import AICSImage

from instanseg.utils.augmentations import Augmentations
from instanseg.utils.utils import _choose_device, show_images
from instanseg.utils.model_loader import load_model


def set_export_paths():
    from pathlib import Path
    if os.environ.get('INSTANSEG_BIOIMAGEIO_PATH'):
        path = Path(os.environ['INSTANSEG_BIOIMAGEIO_PATH'])
    else:
        path = Path(os.path.join(os.path.dirname(__file__),"../bioimageio_models/"))
        os.environ['INSTANSEG_BIOIMAGEIO_PATH'] = str(path)

    if not path.exists():
        path.mkdir(exist_ok=True,parents=True)

    if os.environ.get('INSTANSEG_TORCHSCRIPT_PATH'):
        path = Path(os.environ['INSTANSEG_TORCHSCRIPT_PATH'])
    else:
        path = Path(os.path.join(os.path.dirname(__file__),"../torchscripts/"))
        os.environ['INSTANSEG_TORCHSCRIPT_PATH'] = str(path)

    if not path.exists():
        path.mkdir(exist_ok=True,parents=True)

    if os.environ.get('INSTANSEG_MODEL_PATH'):
        path = Path(os.environ['INSTANSEG_MODEL_PATH'])
    else:
        path = Path(os.path.join(os.path.dirname(__file__),"../models/"))
        os.environ['INSTANSEG_MODEL_PATH'] = str(path)

    if not path.exists():
        path.mkdir(exist_ok=True,parents=True)


dataset_dict = {
    "DSB_2018": ["CC 0","https://bbbc.broadinstitute.org/BBBC038"],
    "CoNSeP": ["Apache 2.0","https://warwick.ac.uk/fac/cross_fac/tia/data/hovernet"],
    "TNBC_2018": ["CC BY 4.0","https://zenodo.org/records/3552674"],
    "MoNuSeg": ["CC BY NC 4.0","https://monuseg.grand-challenge.org/"],
    "LyNSec": ["CC BY 4.0","https://zenodo.org/records/8065174"],
    "LyNSeC": ["CC BY 4.0","https://zenodo.org/records/8065174"],
    "NuInsSeg": ["CC BY 4.0","https://zenodo.org/records/10518968"],
    "IHC_TMA": ["CC BY 4.0","https://zenodo.org/records/7647846"],
    "CPDMI_2023": ["CC BY 4.0","https://www.nature.com/articles/s41597-023-02108-z"],
    "cellpose": ["NC","https://www.cellpose.org/dataset"],
    "TissueNet": ["Modified Apache, Non-Commercial", "https://datasets.deepcell.org/"],
    "CIL": ["CC BY 3.0", "https://www.cellimagelibrary.org/images/CCDB_6843"],
    "BSST265": ["CC0","https://www.ebi.ac.uk/biostudies/bioimages/studies/S-BSST265"],
}



from bioimageio.core.build_spec import build_model
from bioimageio.core.resource_tests import test_model

def readme(model_name: str, model_dict: dict = None):
    # create markdown documentation for your model
    # this should describe how the model was trained, (and on which data)
    # and also what to take into consideration when running the model, especially how to validate the model
    # here, we just create a stub documentation

    with open(os.path.join(model_name, model_name + "_README.md"), "w") as f:
        f.write("# This is an InstanSeg model. \n")
        f.write("The InstanSeg method is shared with an Apache-2.0 license.\n\n")
        f.write("""For an introduction & comparison to other approaches for nucleus segmentation in brightfield histology images, see: \n > Goldsborough, T. et al. (2024) InstanSeg: an embedding-based instance segmentation algorithm optimized for accurate, efficient and portable cell segmentation. _arXiv_. Available at: https://doi.org/10.48550/arXiv.2408.15954. \n\n To read about InstanSeg's extension to nucleus + full cell segmentation and support for fluorescence & multiplexed images, see: \n > Goldsborough, T. et al. (2024) A novel channel invariant architecture for the segmentation of cells and nuclei in multiplexed images using InstanSeg. _bioRxiv_, p. 2024.09.04.611150. Available at: https://doi.org/10.1101/2024.09.04.611150. \n""")
        f.write("\n This model was trained on the following datasets: \n")
        
        if model_dict is not None and "source_dataset" in model_dict.keys():
            for dataset in (model_dict["source_dataset"]).replace("[","").replace("]","").replace("'","").split(", "):
                f.write(f"- {dataset} \n")
                f.write(f"  - License: {dataset_dict[dataset][0]} \n")
                f.write(f"  - URL: {dataset_dict[dataset][1]} \n")

        f.write("\n The user is responsible for ensuring that the model is used in accordance with the licenses of the source datasets. \n")

          #  f.write(str(model_dict["source_dataset"]))


def modify_yaml_for_qupath_config(yaml_path, pixel_size: float, dim_in: int = 3, dim_out: int = 2):

    #copy ijm files
    import shutil
    shutil.copyfile(os.path.join(os.path.dirname(__file__),"./rdf_scripts/instanseg_preprocess.ijm"), os.path.join(os.path.dirname(yaml_path), "instanseg_preprocess.ijm"))
    shutil.copyfile(os.path.join(os.path.dirname(__file__),"./rdf_scripts/instanseg_postprocess.ijm"), os.path.join(os.path.dirname(yaml_path), "instanseg_postprocess.ijm"))

    import yaml
    with open(yaml_path, 'r') as file:
        data = yaml.safe_load(file)

    # Modify the YAML data as needed
    data['config']['qupath'] = {
        'axes': [
            {'role': 'x', 'step': pixel_size, 'unit': 'um'},
            {'role': 'y', 'step': pixel_size, 'unit': 'um'}
        ]}

    data['config']['deepimagej'] = {
    'allow_tiling': True,
    'model_keys': None,
    'prediction': {
        'preprocess': [
            {'kwargs': 'instanseg_preprocess.ijm'}
        ],
        'postprocess': [
            {'kwargs': 'instanseg_postprocess.ijm'}
        ]
    },
    'pyramidal_model': False,
    'test_information': {
        'inputs': [
            {
                'name': 'test-input.npy',
                'pixel_size': {
                    'x': 1.0,
                    'y': 1.0,
                    'z': 1.0
                },
                'size': f'256 x 256 x 1 x {dim_in}'
            }
        ],
        'memory_peak': None,
        'outputs': [
            {
                'name': 'test-output.npy',
                'size': f'256 x 256 x 1 x {dim_out}',
                'type': 'image'
            }
        ],
        'runtime': None
    }

    }

    with open(yaml_path, 'w') as file:
        yaml.dump(data, file)

import os, shutil
def make_archive(source, destination):
        base = os.path.basename(destination)
        name = base.split('.')[0]
        format = base.split('.')[1]
        archive_from = os.path.dirname(source)
        archive_to = os.path.basename(source.strip(os.sep))
        shutil.make_archive(name, format, archive_from, archive_to)
        shutil.move('%s.%s'%(name,format), destination)



def export_bioimageio(torchsript: torch.jit._script.RecursiveScriptModule, 
                      model_name: str, 
                      test_img_path: str, 
                      model_dict: dict = None, 
                      output_name = None):
    
    set_export_paths()

    output_path = os.environ['INSTANSEG_BIOIMAGEIO_PATH']


    if output_name is None:
        output_name = model_name
    # create a directory to store bioimage.io model files
    os.makedirs(output_name, exist_ok=True)
    # save the model weights
    torchsript.save(os.path.join(output_name, "instanseg.pt"))

    model_pixel_size = torchsript.pixel_size
    print("Model pixel size: ", model_pixel_size)


    try:
        model,model_dict = load_model(model_name, path = os.environ['INSTANSEG_MODEL_PATH'])
    except:
        raise Exception("Model configuration files could not be loaded")

    torchsript.eval()
    device = _choose_device()
    torchsript.to(device)

    img = AICSImage(test_img_path)
    if "S" in img.dims.order and img.dims.S > img.dims.C:
        input_data = img.get_image_data("SYX")
    else:
        input_data = img.get_image_data("CYX")
        
    if img.physical_pixel_sizes.X is not None:
        pixel_size = img.physical_pixel_sizes.X
        print("Pixel size was found in the metadata, pixel size is set to: ", pixel_size)
    else:
        pixel_size = 0.5
        print("Pixel size was not found in the metadata, please set the pixel size of the input image in microns manually")

    if model_dict["channel_invariant"]:
        dim_in = 1
        step = 1
    else:
        dim_in = model_dict["dim_in"]
        step = 0

    Augmenter=Augmentations()

    input_tensor,_ = Augmenter.to_tensor(input_data,normalize=False) #this converts the input data to a tensor and does percentile normalization (no clipping)
    input_tensor,_ = Augmenter.normalize(input_tensor, percentile=0.)
    import math
    if math.isnan(model_pixel_size):
        model_pixel_size_tmp = pixel_size
    else:
        model_pixel_size_tmp = model_pixel_size
    input_crop,_ = Augmenter.torch_rescale(input_tensor,labels=None,current_pixel_size=pixel_size,requested_pixel_size=model_pixel_size_tmp,crop = True, random_seed=1)
    input_crop = input_crop.unsqueeze(0) # add batch dimension
    if input_crop.shape[1] != dim_in and not model_dict["channel_invariant"]:
        #input_crop = input_crop[:,1:2,:,:]
        input_crop = torch.zeros((1,dim_in,input_crop.shape[2],input_crop.shape[3]),dtype=torch.float32, device = input_crop.device)

    print("Input tensor shape: ", input_crop.shape)

    # create test data for this model: an input image and an output image
    # this data will be used for model test runs to ensure the model runs correctly and that the expected output can be reproduced
    # NOTE: if you have pre-and-post-processing in your model (see the more advanced models for an example)
    # you will need to save the input BEFORE preprocessing and the output AFTER postprocessing

    np.save(os.path.join(output_name, "test-input.npy"), input_crop.numpy())

    with torch.no_grad():
        output = torchsript(input_crop.to(device))
        dim_out = output.shape[1]
    np.save(os.path.join(output_name, "test-output.npy"), output.cpu().numpy())

    from instanseg.utils.utils import display_overlay

    cover = display_overlay(input_crop[0], output)
    show_images(cover, colorbar=False)
    show_images(cover, colorbar=False, save_str= os.path.join(output_name, "cover"))

    if model_dict is not None and "source_dataset" in model_dict.keys():
        train_data = str(model_dict["source_dataset"])
    else:
        train_data = "Not specified"


    # create readme
    readme(output_name, model_dict)

    if not os.path.exists(output_path):
        os.makedirs(output_path)


    if os.path.exists(os.path.join(output_path, output_name + ".zip")):
        print("Removing existing model zip")
        os.remove(os.path.join(output_path, output_name + ".zip"))

    # now we can use the build_model function to create the zipped package.
    # it takes the path to the weights and data we have just created, as well as additional information
    # that will be used to add metadata to the rdf.yaml file in the model zip
    # we only use a subset of the available options here, please refer to the advanced examples and to the
    # function signature of build_model in order to get an overview of the full functionality

    preprocessing = [[{"name": "scale_range", "kwargs": {"min_percentile": 0.1, "max_percentile": 99.9, "eps": 1e-6, "axes": "xy", "mode": "per_sample"}}]]

    _ = build_model(
        # the weight file and the type of the weights
        weight_uri = os.path.join(output_name, "instanseg.pt"),
        weight_type = "torchscript",
        # the test input and output data
        covers = [os.path.join(output_name, "cover.png")],
        test_inputs = [os.path.join(output_name, "test-input.npy")],
        test_outputs = [os.path.join(output_name, "test-output.npy")],
        # where to save the model zip, how to call the model and a short description of it
        output_path = str(os.path.join(output_path, output_name + ".zip")),
        name = output_name,
        description = "InstanSeg model",
        # additional metadata about authors, licenses, citation etc.
        authors = [{"name": "Thibaut Goldsborough"}],
        license = "Apache-2.0",
        documentation = os.path.join(output_name, output_name + "_README.md"),
        tags = ["cell-segmentation","nuclei","cells","unet","fiji","qupath","pytorch","instanseg","whole-slide-imaging"],  # the tags are used to make models more findable on the website
        cite = [{"text": 
                 "Goldsborough, T. et al. (2024) InstanSeg: an embedding-based instance segmentation algorithm optimized for accurate, efficient and portable cell segmentation. _arXiv_. Available at: https://doi.org/10.48550/arXiv.2408.15954",
                 "doi": "https://doi.org/10.48550/arXiv.2408.15954"},
                {"text": 
                "Goldsborough, T. et al. (2024) A novel channel invariant architecture for the segmentation of cells and nuclei in multiplexed images using InstanSeg. _bioRxiv_, p. 2024.09.04.611150. Available at: https://doi.org/10.1101/2024.09.04.611150.",
                "doi": "https://doi.org/10.1101/2024.09.04.611150"}],
        # description of the tensors
        # these are passed as list because we support multiple inputs / outputs per model
        input_names = ["raw"],
        input_axes = ["bcyx"],
        input_min_shape = [[1, dim_in, 128, 128]],
        input_step = [[0, step, 32, 32]],
        output_names = ["instance"],
        output_axes=["bcyx"],
        output_reference = ["raw"],
        output_scale = [[1.0, 0.0, 1.0, 1.0]],
        output_offset = [[0.0, dim_out /2, 0.0, 0.0]],
        git_repo = "https://github.com/instanseg/instanseg",

        preprocessing = preprocessing,

        pytorch_version = "2.0.0", #str(torch.__version__),
        add_deepimagej_config = True,
      #  pixel_sizes = [{"x":float(model_pixel_size),"y":float(model_pixel_size)}],
    )

    # finally, we test that the expected outptus are reproduced when running the model.
    # the 'test_model' function runs this test.
    # it will output a list of dictionaries. each dict gives the status of a different test that is being run
    # if all of them contain "status": "passed" then all tests were successful
    my_model = bioimageio.core.load_resource_description(os.path.join(output_path, output_name + ".zip")) 
    test_model(my_model)

    #Cleanup 

    files_to_remove = [
        "cover.png",
        "test-output.npy",
        "test-input.npy",
        "instanseg.pt",
        output_name + "_README.md",
        "sample_output_0.tif",
        "sample_input_0.tif"
    ]

    for file in files_to_remove:
        if os.path.exists(file):
            os.remove(file)

    #unzip the folder

    
    import zipfile
    import shutil

    input = os.path.join(output_path, output_name + ".zip")
    destination = os.path.join(output_path, output_name)
    with zipfile.ZipFile(input, 'r') as zip_ref:
        zip_ref.extractall(destination)
    
    yaml_path = os.path.join(destination, 'rdf.yaml')
    modify_yaml_for_qupath_config(yaml_path, pixel_size=model_pixel_size, dim_in=dim_in, dim_out=dim_out)

    
    make_archive(destination, input)


    

    

    shutil.rmtree(output_name)



import torch
import torch
import torchvision.transforms.functional as TF
import sys
import torch.nn as nn
import pdb
import numpy as np
import matplotlib.pyplot as plt
from tqdm.auto import tqdm
from instanseg.utils.metrics import _robust_average_precision, _robust_f1_mean_calculator


from instanseg.utils.augmentations import Augmentations
from instanseg.utils.utils import timer
import time
import torchvision

from instanseg.utils.utils import show_images
import warnings


global_step = 0
def train_epoch(train_model, 
                train_device, 
                train_dataloader, 
                train_loss_fn, 
                train_optimizer, 
                args,
                writer = None,
                ):
    
    global global_step
    start = time.time()
    train_model.train()
    train_loss = []
    for image_batch, labels_batch, _ in tqdm(train_dataloader, disable=args.on_cluster):

        image_batch = image_batch.to(train_device)
        labels = labels_batch.to(train_device)
        output = train_model(image_batch)
        loss = train_loss_fn(output, labels.clone()).mean()
        train_optimizer.zero_grad()
        loss.backward()

        torch.nn.utils.clip_grad_norm_(train_model.parameters(), args.clip)

        if writer is not None:
            writer.add_scalar('train_loss', loss, global_step)
            global_step += 1

        train_optimizer.step()
        train_loss.append(loss.detach().cpu().numpy())

    end = time.time()

    return np.mean(train_loss), end - start
    

global_step_test = 0
def test_epoch(test_model, 
               test_device, 
               test_dataloader, 
               test_loss_fn, 
               args,
               postprocessing_fn,
               method,
               iou_threshold,
               debug=False, 
               save_str=None, 
               save_bool=False,
               writer = None,
               best_f1=None):
    global global_step_test
    start = time.time()

    test_model.eval()
    test_loss = []

    current_f1_list = []
    with torch.no_grad():
        for image_batch, labels_batch, _ in tqdm(test_dataloader, disable=args.on_cluster):
            image_batch = image_batch.to(test_device)
            labels = labels_batch.to(test_device) 
            output = test_model(image_batch)  
            loss = test_loss_fn(output, labels.clone()).mean()
            test_loss.append(loss.detach().cpu().numpy())
            writer.add_scalar('test_loss', loss, global_step_test)


            if labels.type() != 'torch.cuda.FloatTensor' and labels.type() != 'torch.FloatTensor':
                predicted_labels = torch.stack([postprocessing_fn(out) for out in output])
                f1i = _robust_average_precision(labels.clone(), predicted_labels.clone(),
                                               threshold=iou_threshold)

                current_f1_list.append((f1i))
            else:
                warnings.warn("Labels are of type float, not int. Not calculating F1.")
                current_f1_list.append(0)

            if global_step_test % 100 == 0:
                img = torchvision.utils.make_grid(output[0].unsqueeze(1), scale_each=True, normalize=True, nrow = method.dim_out //2 if args.cells_and_nuclei else method.dim_out )
                writer.add_images('Model Outputs', img[None], global_step)

            if args.channel_invariant and global_step_test % 100 == 0:
                img = torchvision.utils.make_grid(test_model.model.AdaptorNet(image_batch),scale_each = True, normalize = True)
                writer.add_images('ChIN rendering', img[None], global_step)
            if args.cells_and_nuclei:
                writer.add_scalar('test_f1_nuclei',current_f1_list[-1][0], global_step_test)
                writer.add_scalar('test_f1_cells', current_f1_list[-1][1], global_step_test)
            else:
                writer.add_scalar('test_f1', current_f1_list[-1], global_step_test)
            global_step_test += 1

    f1_array = np.array(current_f1_list)  # either N,2 or N,

    if f1_array.ndim == 1:
        f1_array = np.atleast_2d(f1_array).T

    mean1_f1 = np.nanmean(f1_array, axis=0)

    mean_f1 = _robust_f1_mean_calculator(mean1_f1)
    #  mean_f1 = current_f1_list

    if mean_f1 > best_f1 or save_bool:
        if len(image_batch[0]) == 3:
            input1 = image_batch[0]
        else:
            input1 = image_batch[0][0]
        labels_dst = labels[0]
        lab = postprocessing_fn(output[0])

        if lab.squeeze().dim() == 2:
            show_images([input1] + [label_i for label_i in labels_dst] + [lab] + [out for out in output[0]],
                        save_str=save_str,
                        titles=["Source"] + ["Label" for _ in labels_dst] + ["Prediction"] + ["Out" for _ in output[0]],
                        labels=[1, 2])
        else:
            show_images([input1] + [label_i for label_i in labels_dst] + [label_i for label_i in lab] + [out for out in
                                                                                                         output[0]],
                        save_str=save_str,
                        titles=["Source"] + ["Label: Nuclei", "Label: Cells"] + ["Prediction: Nuclei",
                                                                                 "Prediction: Cells"] + ["Out" for _ in
                                                                                                         output[0]],
                        labels=[1, 2, 3, 4], n_cols=5)

    end = time.time()
    return np.mean(test_loss), mean1_f1, end - start


def collate_fn(data):
    # data is of length batch size
    # data[0][0] is first image, data[0][1] os the first label

    # print(data[0][0].shape,len(data))
    """
       data: is a list of tuples with (example, label, length)
             where 'example' is a tensor of arbitrary shape
             and label/length are scalars
    """
    imgs, labels = zip(*data)
    lengths = [img.shape[0] for img in imgs]

    max_len = max(lengths)
    C, H, W = data[0][0].shape
    images = torch.zeros((len(data), max_len, H, W))
    labels = torch.stack(labels)
    lengths = torch.tensor(lengths)

    for i, img in enumerate(imgs):
        images[i, :len(img)] = img

    return images, labels, lengths.int()


# import fastremap
class Segmentation_Dataset():
    def __init__(self, img, label, common_transforms=True, metadata=None, size=(256, 256), augmentation_dict=None,
                 dim_in=3, debug=False, cells_and_nuclei=False, target_segmentation="N", channel_invariant = False):
        self.X = img
        self.Y = label
        self.common_transforms = common_transforms

        assert len(self.X) == len(self.Y), "The number of images and labels must be the same"
        if len(metadata) == 0:
            self.metadata = [None] * len(self.X)
        else:
            self.metadata = metadata

        assert len(self.X) == len(self.metadata), print("The number of images and metadata must be the same")
        self.size = size
        self.Augmenter = Augmentations(augmentation_dict=augmentation_dict, debug=debug, shape=self.size,
                                       dim_in=dim_in, cells_and_nuclei=cells_and_nuclei,
                                       target_segmentation=target_segmentation, channel_invariant = channel_invariant)

    def __len__(self):
        return len(self.X)

    def __getitem__(self, i):

        data = self.X[i]
        label = self.Y[i]
        meta = self.metadata[i]

        if self.common_transforms:
            data, label = self.Augmenter(data, label, meta)

        if len(label.shape) == 2:
            label = label[None, :]
        if len(data.shape) == 2:
            data = data[None, :]

        assert not data.isnan().any(), "Tranformed images contains NaN"
        assert not label.isnan().any(), "Transformed labels contains NaN"

        return data.float(), label



def plot_loss(_model):
    loss_fig = plt.figure()
    timer = loss_fig.canvas.new_timer(interval=300000)
    timer.add_callback(plt.close)

    losses = [param.grad.norm().item() for name, param in _model.named_parameters() if param.grad is not None]
    names = [name for name, param in _model.named_parameters() if param.grad is not None]

    plt.plot(losses)
    plt.xticks(np.arange(len(names))[::1], names[::1])
    plt.xticks(fontsize=8, rotation=90)
    spacing = 0.5
    loss_fig.subplots_adjust(bottom=spacing)
    timer.start()
    plt.show()



def check_max_grad(_model):
    losses = np.array([param.grad.norm().item() for name, param in _model.named_parameters() if param.grad is not None])
    return losses.max()


def check_min_grad(_model):
    losses = np.array([param.grad.norm().item() for name, param in _model.named_parameters() if param.grad is not None])
    return losses.min()


def check_mean_grad(_model):
    losses = np.array([param.grad.norm().item() for name, param in _model.named_parameters() if param.grad is not None])
    return losses.mean()



def optimize_hyperparameters(model,postprocessing_fn, data_loader = None, val_images = None, val_labels = None,max_evals = 50, verbose = False, threshold = [0.5, 0.7, 0.9], show_progressbar = True, device = None):


    from instanseg.utils.metrics import _robust_average_precision
    from instanseg.utils.utils import _choose_device

    from hyperopt import fmin
    from hyperopt import hp
    from hyperopt import Trials
    from hyperopt import tpe
    import copy

    if device is None:
        device = _choose_device()

    bayes_trials = Trials()

    space = {  # instanseg
        'mask_threshold': hp.uniform('mask_threshold', 0.3, 0.7),
        'seed_threshold': hp.uniform('seed_threshold', 0.5, 1),
        'overlap_threshold': hp.uniform('overlap_threshold', 0.1, 0.9),
        #'min_size': hp.uniform('min_size', 0, 30),
        'peak_distance': hp.uniform('peak_distance', 3, 10),
        'mean_threshold': hp.uniform('mean_threshold', 0.0, 0.3)} #the max could be increased, but may cuase the method not to converge for some reason.
    
    _model = model # copy.deepcopy(model)
    _model.eval()
    predictions = []

    with torch.no_grad():
        if data_loader is not None:
            for image_batch, labels_batch, _ in data_loader:
                    image_batch = image_batch.to(device)
                    output = _model(image_batch).cpu()
                    predictions.extend([pred,masks] for pred,masks in zip(output,labels_batch))


            def objective(params={}):
                pred_masks = []
                gt_masks = []
                for pred, masks in predictions:
                    lab = postprocessing_fn(pred.to(device), **params).cpu()
                    pred_masks.append(lab)
                    gt_masks.append(masks)

                mean_f1 = _robust_average_precision(torch.stack(gt_masks),torch.stack(pred_masks),threshold = threshold)

                if type(mean_f1) == list:
                    mean_f1 = np.nanmean(mean_f1)

                return 1 - mean_f1
        
        elif val_images is not None and val_labels is not None:
            from instanseg.utils.tiling import instanseg_padding, recover_padding
            def objective(params={}):
                pred_masks = []
                gt_masks = []
                #randomly shuffle val_images and val_labels

                np.random.seed(0)
                indexes = np.random.permutation(len(val_images))[:300]
                indexes.sort()

                for i in indexes:
                    imgs = val_images[i]
                    gt_mask = val_labels[i]
                    with torch.no_grad():
                        imgs = imgs.to(device)
                        imgs, pad = instanseg_padding(imgs, min_dim = 32)
                        output = _model(imgs[None,])
                        output = recover_padding(output, pad).squeeze(0)
                        lab = postprocessing_fn(output.to(device), **params).cpu()
                        pred_masks.append(lab)
                        gt_masks.append(gt_mask)

                mean_f1 = _robust_average_precision(gt_masks,pred_masks,threshold = threshold)

                if type(mean_f1) == list:
                    mean_f1 = np.nanmean(mean_f1)

                return 1 - mean_f1
        else:
            raise ValueError("Either data_loader or val_images and val_labels must be provided")

        print("Optimizing hyperparameters")
        # Optimize
        best = fmin(fn=objective, space=space, algo=tpe.suggest,
                    max_evals=max_evals, trials=bayes_trials, show_progressbar = show_progressbar)
    
    if verbose:
        print(best)
    return best






import pdb
import torch.nn as nn
import torch
import monai

import torch.distributions as td
import torch.nn as nn

import torch.nn as nn
import torch
import torch.distributions as td
import numpy as np

from monai.networks.blocks.convolutions import Convolution, ResidualUnit



def build_monai_model(model_str: str, build_model_dictionary: dict):


    if model_str == "AttentionUNet":
        from monai.networks.nets import AttentionUnet

        model = AttentionUnet(spatial_dims=2, in_channels=int(build_model_dictionary["dim_in"]),
                              out_channels=build_model_dictionary["dim_out"], \
                              dropout=build_model_dictionary["dropprob"], channels=build_model_dictionary["layers"], \
                              strides=tuple([2 for _ in build_model_dictionary["layers"][:-1]])
                              )
    elif model_str == "FlexibleUNet":
        from monai.networks.nets import FlexibleUNet
        model = FlexibleUNet(in_channels=build_model_dictionary["dim_in"],
                             out_channels=build_model_dictionary["dim_out"], dropout=build_model_dictionary["dropprob"],
                             backbone="efficientnet-b0")
        

    elif model_str == "BasicUNetPlusPlus":
        from monai.networks.nets import BasicUNetPlusPlus
        model = BasicUNetPlusPlus(spatial_dims=2, in_channels=build_model_dictionary["dim_in"],
                                  out_channels=build_model_dictionary["dim_out"],
                                  dropout=build_model_dictionary["dropprob"])

        class ModelWrapper(BasicUNetPlusPlus):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def forward(self, inputs):
                output = super().forward(inputs)
                modified_output = output[0]  # Modify the output here as needed
                return modified_output

        model = ModelWrapper(spatial_dims=2, in_channels=build_model_dictionary["dim_in"],
                             out_channels=build_model_dictionary["dim_out"], dropout=build_model_dictionary["dropprob"])

    elif model_str == "UNETR":
        from monai.networks.nets import UNETR
        model = UNETR(in_channels=build_model_dictionary["dim_in"], out_channels=build_model_dictionary["dim_out"],
                      img_size=256, feature_size=32, norm_name='batch', spatial_dims=2)


    else:
        raise NotImplementedError("Model not implemented: " + model_str)

    return model


def read_model_args_from_csv(path=r"../results/", folder=""):
    import pandas as pd
    from pathlib import Path
    model_path = Path(path) / folder
    df = pd.read_csv(model_path / "experiment_log.csv", header=None)
    build_model_dictionary = dict(zip(list(df[0]), list(df[1])))

    if "model_shape" in build_model_dictionary.keys():
        build_model_dictionary["model_shape"] = eval(build_model_dictionary["model_shape"])
    for key in ["dim_in", "n_sigma", "dim_out", "dim_coords"]:
        build_model_dictionary[key] = eval(str(build_model_dictionary[key])) if str(
            build_model_dictionary[key]) != "nan" else None
    if "to_centre" in build_model_dictionary.keys():
        build_model_dictionary["to_centre"] = eval(build_model_dictionary["to_centre"])
    if "dropprob" in build_model_dictionary.keys():
        build_model_dictionary["dropprob"] = float(build_model_dictionary["dropprob"])
    if "layers" in build_model_dictionary.keys():
        build_model_dictionary["layers"] = tuple(eval(build_model_dictionary["layers"]))
    if "requested_pixel_size" in build_model_dictionary.keys():
        build_model_dictionary["pixel_size"] = float(build_model_dictionary["requested_pixel_size"])
    if "cells_and_nuclei" in build_model_dictionary.keys():
        build_model_dictionary["cells_and_nuclei"] = bool(eval(build_model_dictionary["cells_and_nuclei"]))
    if "norm" in build_model_dictionary.keys():
        if build_model_dictionary["norm"] == "None" or str(build_model_dictionary["norm"]).lower() == "nan":
            build_model_dictionary["norm"] = None
        else:
            build_model_dictionary["norm"] = str(build_model_dictionary["norm"])
    else:
        print("Norm not specified in model dictionary")
        build_model_dictionary["norm"] = None
    if "feature_engineering" in build_model_dictionary.keys():
        build_model_dictionary["feature_engineering"] = str(build_model_dictionary["feature_engineering"])
    else:
        print("Feature engineering not specified in model dictionary")
        build_model_dictionary["feature_engineering"] = "0"
    if "adaptor_net_str" in build_model_dictionary.keys():
        build_model_dictionary["adaptor_net_str"] = str(build_model_dictionary["adaptor_net_str"])
    if "multihead" in build_model_dictionary.keys():
        build_model_dictionary["multihead"] = bool(eval(build_model_dictionary["multihead"]))
    else:
        build_model_dictionary["multihead"] = False
    if "channel_invariant" in build_model_dictionary.keys():
        build_model_dictionary["channel_invariant"] = bool(eval(build_model_dictionary["channel_invariant"]))

    return build_model_dictionary


def build_model_from_dict(build_model_dictionary):
    if build_model_dictionary["dim_in"] == 0 or build_model_dictionary["dim_in"] is None:
        dim_in = 3  # Channel invariance currently outputs a 3 channel image
    else:
        dim_in = build_model_dictionary["dim_in"]

    if "dropprob" not in build_model_dictionary.keys():
        build_model_dictionary["dropprob"] = 0.0

    if build_model_dictionary["model_str"] == "InstanSeg_UNet":
            from instanseg.utils.models.InstanSeg_UNet import InstanSeg_UNet
            print("Generating InstanSeg_UNet")
            multihead = build_model_dictionary["multihead"]

            if build_model_dictionary["cells_and_nuclei"]:
                if not multihead:
                    from itertools import chain
                    out_channels = [[build_model_dictionary["dim_coords"], build_model_dictionary["n_sigma"],1] for i in range(2)]
                    out_channels = list(chain(*out_channels))
                
                else:
                    out_channels = [[build_model_dictionary["dim_coords"], build_model_dictionary["n_sigma"],1] for i in range(2)]

            else:
                if not multihead:
                    out_channels = [[build_model_dictionary["dim_coords"], build_model_dictionary["n_sigma"],1]]
                else:
                    out_channels = [[build_model_dictionary["dim_coords"]], [build_model_dictionary["n_sigma"]],[1]]

            model = InstanSeg_UNet(in_channels=dim_in, 
                            layers = np.array(build_model_dictionary["layers"])[::-1],
                            out_channels=out_channels,
                            norm  = build_model_dictionary["norm"], 
                            dropout=build_model_dictionary["dropprob"])
            
    else:
        model = build_monai_model(build_model_dictionary["model_str"], build_model_dictionary)

    return model


def remove_module_prefix_from_dict(dictionary):
    """
    Removes the module prefix from a dictionary of model weights
    :param dictionary: dictionary of model weights
    :return: modified dictionary
    """
    modified_dict = {}
    for key, value in dictionary.items():
        modified_dict[key] = value
    return modified_dict


def has_pixel_classifier_state_dict(state_dict):
    return bool(sum(['pixel_classifier' in key for key in state_dict.keys()]))


def has_adaptor_net_state_dict(state_dict):
    return bool(sum(['AdaptorNet' in key for key in state_dict.keys()]))

def has_pixel_classifier_model(model):
    import torch
    for module in model.modules():
        if isinstance(module, torch.nn.Module):
            module_class = module.__class__.__name__
            if 'pixel_classifier' in module_class or 'ProbabilityNet' in module_class:
                return True
    return False


def load_model_weights(model, device, folder, path=r"../models/", dict = None):
    import torch
    from pathlib import Path
    model_path = Path(path) / folder
    if torch.cuda.is_available():
        model_dict = torch.load(model_path / "model_weights.pth")
    else:
        if device is None:
            if torch.backends.mps.is_available():
                device = 'mps'
                print('CUDA not available - attempting to load MPS model')
            else:
                device = 'cpu'
                print('CUDA not available - attempting to load CPU model')
        model_dict = torch.load(model_path / "model_weights.pth", map_location=device)

    model_dict['model_state_dict'] = remove_module_prefix_from_dict(model_dict['model_state_dict'])

    if has_pixel_classifier_state_dict(model_dict['model_state_dict']) and not has_pixel_classifier_model(model):
        from instanseg.utils.loss.instanseg_loss import InstanSeg

        method = InstanSeg(n_sigma=int(dict["n_sigma"]), feature_engineering_function= dict["feature_engineering"],dim_coords = dict["dim_coords"],device =device)
        model = method.initialize_pixel_classifier(model, MLP_width=int(dict["mlp_width"]))

    from instanseg.utils.models.ChannelInvariantNet import AdaptorNetWrapper, has_AdaptorNet
    if has_adaptor_net_state_dict(model_dict['model_state_dict']) and not has_AdaptorNet(model):
        from instanseg.utils.models.ChannelInvariantNet import AdaptorNetWrapper, has_AdaptorNet
        model = AdaptorNetWrapper(model, norm = dict["norm"],adaptor_net_str = dict["adaptor_net_str"])

    #from instanseg.utils.AI_utils import set_running_stats
    #set_running_stats(model,device = "cuda")

    model.load_state_dict(model_dict['model_state_dict'], strict=True)
    model.to(device)

    return model, model_dict

def load_model(folder,path=r"../models/", device='cpu'):
    build_model_dictionary = read_model_args_from_csv(path=path, folder=folder)

    empty_model = build_model_from_dict(build_model_dictionary)

    model, _ = load_model_weights(empty_model, path=path, folder=folder, device=device, dict = build_model_dictionary)

    return model, build_model_dictionary



import torch
import numpy as np
import pdb

from typing import Tuple, List, Union

from instanseg.utils.loss.lovasz_losses import binary_xloss
from instanseg.utils.pytorch_utils import torch_fastremap, torch_onehot, remap_values, fast_iou, fast_sparse_iou, eccentricity_batch, connected_components
from instanseg.utils.tiling import instanseg_padding, recover_padding

import torch.nn.functional as F
import torch.nn as nn

binary_xloss = torch.nn.BCEWithLogitsLoss()
l1_loss = torch.nn.L1Loss()

from instanseg.utils.utils import show_images

from instanseg.utils.utils import timer

integer_dtype = torch.int64



def convert(prob_input: torch.Tensor, coords_input: torch.Tensor, size: Tuple[int, int],
            mask_threshold: float = 0.5) -> torch.Tensor:
    # Create an array of labels for each pixel
    all_labels = torch.arange(1, 1 + prob_input.shape[0], dtype=torch.float32, device=prob_input.device)
    labels = torch.ones_like(prob_input) * torch.reshape(all_labels, (-1, 1, 1, 1))

    # Get flattened arrays
    labels = labels.flatten()
    prob = prob_input.flatten()
    x = coords_input[0, ...].flatten()
    y = coords_input[1, ...].flatten()

    # Predict image dimensions if we don't have them
    if size is None:
        size = (int(y.max() + 1), int(x.max() + 1))

    # Find indices with above-threshold probability values
    inds_prob = prob >= mask_threshold
    n_thresholded = torch.count_nonzero(inds_prob)
    if n_thresholded == 0:
        return torch.zeros(size, dtype=torch.float32, device=labels.device)

    # Create an array of [linear index, y, x, label], skipping low-probability values
    arr = torch.zeros((int(n_thresholded), 5), dtype=coords_input.dtype, device=labels.device)
    arr[:, 1] = y[inds_prob]
    arr[:, 2] = x[inds_prob]
    # NOTE: UNEXPECTED Y,X ORDER!
    arr[:, 0] = arr[:, 2] * size[1] + arr[:, 1]
    arr[:, 3] = labels[inds_prob]

    # Sort first by descending probability
    inds_sorted = prob[inds_prob].argsort(descending=True, stable=True)
    arr = arr[inds_sorted, :]

    # Stable sort by linear indices
    inds_sorted = arr[:, 0].argsort(descending=False, stable=True)
    arr = arr[inds_sorted, :]

    # Find the first occurrence of each linear index - this should correspond to the label
    # that has the highest probability, because they have previously been sorted
    inds_unique = torch.ones_like(arr[:, 0], dtype=torch.bool)
    inds_unique[1:] = arr[1:, 0] != arr[:-1, 0]

    # Create the output
    output = torch.zeros(size, dtype=torch.float32, device=labels.device)
    # NOTE: UNEXPECTED Y,X ORDER!
    output[arr[inds_unique, 2], arr[inds_unique, 1]] = arr[inds_unique, 3].float()

    return output


def find_all_local_maxima(image: torch.Tensor, neighbourhood_size: int, minimum_value: float) -> torch.Tensor:
    """
        helper function for peak_local_max that finds all the local maxima
        within each neighbourhood. (may return multiple per neighbourhood).
        """
    # Perform max pooling with the specified neighborhood size
    kernel_size = 2 * neighbourhood_size + 1
    pooled = F.max_pool2d(image, kernel_size=kernel_size, stride=1, padding=kernel_size // 2)

    # Create a mask for the local maxima
    mask = (pooled == image) * (image >= minimum_value)

    # Apply the mask to the original image to retain only the local maxima values
    local_maxima = image * mask

    return local_maxima


def torch_peak_local_max(image: torch.Tensor, neighbourhood_size: int, minimum_value: float,
                             return_map: bool = False, dtype: torch.dtype = torch.int) -> torch.Tensor:
    """
    UPDATED FOR PERFORMANCE TESTING - NOT IDENTICAL, AS USES *FIRST* MAX, NOT FURTHEST FROM ORIGIN
    """
    h, w = image.shape
    image = image.view(1, 1, h, w)
    device = image.device

    kernel_size = 2 * neighbourhood_size + 1
    pooled, max_inds = F.max_pool2d(image, kernel_size=kernel_size, stride=1, padding=neighbourhood_size,
                                    return_indices=True)

    inds = torch.arange(0, image.numel(), device=device, dtype=dtype).reshape(image.shape)

    peak_local_max = (max_inds == inds) * (pooled > minimum_value)

    if return_map:
        return peak_local_max

    # Non-zero causes host-device synchronization, which is a bottleneck
    return torch.nonzero(peak_local_max.squeeze()).to(dtype)

def torch_peak_local_max_LEGACY(image: torch.Tensor, neighbourhood_size: int, minimum_value: float, return_map: bool = False) -> torch.Tensor:
    """
    computes peak local maxima function for an image (or batch of images), returning a maxima mask
    and the coordinates of the peak local max values.
    peak local maxima returns a image that is zero at all points other than local maxima.
    At the local maxima, the pixel retains its value in the original image.
    
    image: a torch tensor of shape [B,1,H,W] or [H,W], B is batch size. H,W are spatial dims.
    neighbourhood_size: int. Only one maximum will be selected within a square patch of width
        equal to the neighbourhood size (specifically the largest maxima in that neighbourhood).
        Where there are multiple local maxima with the largest value within the neighbourhood,
        the maxima furthest away from the origin (furthest by euclidian distance from pixel (0,0))
        is retained (ensuring there is only one maximum per neighbourhood).
    minimum_value: float. Local maxima with pixel intensity below this value are ignored.
    
    returns: a torch tensor of shape equal to image, a list of length B containing (lx, ly) pairs
    where lx and ly are torch tensors containing the x and y coordinates of each local maxima for a given image.
    if image has shape [H,W], returns (lx, ly). 
    """
    assert image.ndim == 2, "image must be of shape [H,W]"

    h, w = image.shape
    image = image.view(1, 1, h, w)
    device = image.device

    all_local_maxima = find_all_local_maxima(image, neighbourhood_size, minimum_value)

    # perform non-maximal coordinate suppression to only get one maximum per neighbourhood.
    # specifically, where there are two maxima in a neighbourhood, I retain the maxima
    # which has the furthest euclidian distance away from the origin. This is just an
    # 'arbitrary' way for me to split the ties. 
    spatial_dims = [image.shape[-2], image.shape[-1]]

    grid = torch.stack(
        torch.meshgrid(
            torch.arange(0, spatial_dims[0], 1, device=device, dtype = torch.float32), torch.arange(0, spatial_dims[1], 1, device=device, dtype = torch.float32),
            indexing='ij'
        )
    )

    distance_to_origin = (grid.unsqueeze(0)).square().sum(dim=1).sqrt()

    distance_of_max_poses = torch.mul(all_local_maxima, distance_to_origin)

    retained_maxima = find_all_local_maxima(distance_of_max_poses, neighbourhood_size, minimum_value=minimum_value)
    peak_local_max = all_local_maxima * (retained_maxima > minimum_value)

    if return_map:
        return peak_local_max

    locs = grid[:,peak_local_max.squeeze()>0].T.int()


    return locs


def centre_crop(centroids: torch.Tensor, window_size: int, h:int, w:int) -> torch.Tensor:
    """
    centres the crop around the centroid, ensuring that the crop does not exceed the image dimensions.
    """
    C = centroids.shape[0]
    centroids = centroids.clone()  # C,2
    centroids[:,0] = centroids[:, 0].clamp(min=window_size //2 , max=h - window_size //2)
    centroids[:, 1] = centroids[:,1].clamp(min=window_size //2, max=w - window_size //2)
    window_slices = (centroids[:, None] + torch.tensor([[-1, -1], [1, 1]], device = centroids.device) * (window_size //2))


    grid_x, grid_y = torch.meshgrid(
        torch.arange(window_size, device=centroids.device, dtype=centroids.dtype),
        torch.arange(window_size, device=centroids.device, dtype=centroids.dtype), indexing="ij")
    
    mesh = torch.stack((grid_x, grid_y))

    mesh_grid = mesh.expand(C, 2, window_size, window_size)  # C,2,2*window_size,2*window_size
    mesh_grid_flat = torch.flatten(mesh_grid, 2).permute(1, 0, -1)  # 2,C,2*window_size*2*window_size
    idx = window_slices[:, 0].permute(1, 0)[:, :, None]
    mesh_grid_flat = mesh_grid_flat + idx
    mesh_grid_flat = torch.flatten(mesh_grid_flat, 1)  # 2,C*2*window_size*2*window_size

    return mesh_grid_flat


def compute_crops( x: torch.Tensor, 
                  c: torch.Tensor, 
                  sigma: torch.Tensor,
                  centroids_idx: torch.Tensor,
                  feature_engineering, 
                  pixel_classifier, 
                  window_size: int = 128):

    h, w = x.shape[-2:]
    C = c.shape[0]

    mesh_grid_flat = centre_crop(centroids_idx, window_size, h, w)

    x = feature_engineering(x, c, sigma, window_size //2 , mesh_grid_flat)


    x = pixel_classifier(x)  # C*H*W,1

    x = x.view(C, 1, window_size, window_size)

    idx = torch.arange(1, C + 1, device=x.device, dtype = mesh_grid_flat.dtype)

    rep = torch.ones((C, window_size, window_size), device=x.device, dtype =mesh_grid_flat.dtype)
    rep = rep * (idx[:, None, None] - 1)

    iidd = torch.cat((rep.flatten()[None,], mesh_grid_flat)).to(mesh_grid_flat.dtype)

    return x, iidd



def find_connected_components(adjacency_matrix: torch.Tensor):

    M = (adjacency_matrix + torch.eye(adjacency_matrix.shape[0],
                               device=adjacency_matrix.device))  # https://math.stackexchange.com/questions/1106870/can-i-find-the-connected-components-of-a-graph-using-matrix-operations-on-the-gr
    num_iterations = 10
    out = torch.matrix_power(M, num_iterations)
    col = torch.arange(0, out.shape[0], device=out.device).view(-1, 1).expand(out.shape[0], out.shape[
        0])  # Just a column matrix with numbers from 0 to out.shape[0]
    out_col_idx = ((out > 1).int() - torch.eye(out.shape[0], device=out.device)) * col
    maxes = out_col_idx.argmax(0) * (out_col_idx.max(0)[0] > 0).int()
    maxes = torch.maximum(maxes + 1, (torch.arange(0, out.shape[0],
                                                    device=out.device) + 1))  # recover the diagonal elements that were suppressed
    tentative_remapping = torch.stack(((torch.arange(0, out.shape[0], device=out.device) + 1), maxes))
    # start with two zeros:
    remapping = torch.cat((torch.zeros(2, 1, device=tentative_remapping.device), tentative_remapping),
                            dim=1)  # Maybe this can be avoided in the future by thresholding labels
    
    return remapping
    


def has_pixel_classifier_model(model):
    for module in model.modules():
        if isinstance(module, torch.nn.Module):
            module_class = module.__class__.__name__
            if 'ProbabilityNet' in module_class:
                return True
    return False




def merge_sparse_predictions(x: torch.Tensor, coords: torch.Tensor, mask_map: torch.Tensor, size : list[int], mask_threshold: float = 0.5, window_size = 128, min_size = 10, overlap_threshold = 0.5, mean_threshold = 0.5):

    # top_left = window_slices[:,0,:]
    # shifted_centroid = centroids_idx - top_left
    # cc = connected_components((output>mask_threshold).float())
    # labels_to_keep = cc[torch.arange(cc.shape[0]),0,shifted_centroid[:,0],shifted_centroid[:,1]]
    # in_mask = cc == labels_to_keep[:,None,None,None]
    # output *= in_mask


    labels = convert(x, coords, size=(size[1], size[2]), mask_threshold=mask_threshold)[None]

    idx = torch.arange(1, size[0] + 1, device=x.device, dtype =coords.dtype)
    stack_ID = torch.ones((size[0], window_size, window_size), device=x.device, dtype=coords.dtype)
    stack_ID = stack_ID * (idx[:, None, None] - 1)

    coords = torch.stack((stack_ID.flatten(), coords[0] * size[2] + coords[1])).to(coords.dtype)

    fg = x.flatten() > mask_threshold
    x = x.flatten()[fg]
    coords = coords[:, fg]

    using_mps = False
    if x.is_mps:
        using_mps = True
        device = 'cpu'
        x = x.to(device)
        mask_map = mask_map.to(device)

    sparse_onehot = torch.sparse_coo_tensor(
        coords,
        x.flatten() > mask_threshold,
        size=(size[0], size[1] * size[2]),
        dtype=x.dtype,
        device=x.device,
        requires_grad=False,
        )

    object_areas = torch.sparse.sum(sparse_onehot, dim=1).values()

    sum_mask_value = torch.sparse.sum((sparse_onehot * mask_map.flatten()[None]), dim=1).values()
    mean_mask_value = sum_mask_value / object_areas
    objects_to_remove = ~torch.logical_and(mean_mask_value > mean_threshold, object_areas > min_size)

    if window_size **2 * sparse_onehot.shape[0] == sparse_onehot.sum():
        #This can happen at the start of training. This can cause OOM errors and is never a good sign - may aswell abort.
        return labels

    iou = fast_sparse_iou(sparse_onehot)

    remapping = find_connected_components((iou>overlap_threshold).float() )

    if using_mps:
        device = 'mps'
        remapping = remapping.to(device)
        labels = labels.to(device)

    labels = remap_values(remapping, labels)

    labels_to_remove = (torch.arange(0, len(objects_to_remove), device=objects_to_remove.device, dtype = coords.dtype) + 1)[
    objects_to_remove]

    labels[torch.isin(labels, labels_to_remove)] = 0

    return labels

def guide_function(params: torch.Tensor,device ='cuda', width: int = 256):

    #params must be depth,3  

    depth = params.shape[0]
    xx = torch.linspace(0, 1, width, device=device).view(1, 1, -1).expand(1, width,width)
    yy = torch.linspace(0, 1, width, device=device).view(1, -1, 1).expand(1, width, width)
    xxyy  = torch.cat((xx, yy), 0).expand(depth,2,width,width)

    xx = xxyy[:,0] * params[:,0][:,None,None]
    yy = xxyy[:,1] * params[:,1][:,None,None]

    return torch.sin(xx+yy+params[:,2,None,None])[None]


def generate_coordinate_map(mode: str = "linear", spatial_dim: int = 2, height: int = 256, width: int = 256, device: torch.device = torch.device(type='cuda')):

    if mode == "linear":
        if spatial_dim ==2:
            xx = torch.linspace(0, width * 64 / 256, width, device=device).view(1, 1, -1).expand(1, height,width)
            yy = torch.linspace(0, height * 64 / 256, height, device=device).view(1, -1, 1).expand(1, height, width)
            xxyy = torch.cat((xx, yy), 0)

        elif spatial_dim >= 3:
            xx = torch.linspace(0, width * 64 / 256, width, device=device).view(1, 1, -1).expand(1, height,width)
            yy = torch.linspace(0, height * 64 / 256, height, device=device).view(1, -1, 1).expand(1, height, width)
            zz = torch.zeros_like(xx).expand(spatial_dim - 2,-1,-1)
            xxyy = torch.cat((xx, yy,zz), 0)
        else:
            xxyy = torch.zeros((spatial_dim, height, width), device=device) #NOT IMPLEMENTED - THIS IS JUST A DUMMY VALUE

    else:
        xxyy = torch.zeros((spatial_dim, height, width), device=device) #NOT IMPLEMENTED - THIS IS JUST A DUMMY VALUE


    return xxyy



class ProbabilityNet(nn.Module):
    def __init__(self, embedding_dim=4, width = 5):
        super().__init__()
        self.fc1 = nn.Linear(embedding_dim, width)
        self.fc2 = nn.Linear(width, width)
        self.fc3 = nn.Linear(width, 1)

    def forward(self, x):
        # x is C*H*W,E+S+1 (H,W is the window of the crop used here, e.g 100x100, not the image)
      #  with torch.cuda.amp.autocast():
        x = self._relu_non_empty(self.fc1(x))
        x = self._relu_non_empty(self.fc2(x))
        x = self.fc3(x)
        return x

    def _relu_non_empty(self, x: torch.Tensor) -> torch.Tensor:
        """
        Workaround for https://github.com/pytorch/pytorch/issues/118845 on MPS
        """
        if x.numel() == 0:
            return x
        else:
            return torch.relu_(x)



class MyBlock(nn.Sequential):
    def __init__(self, embedding_dim, width):
        super(MyBlock, self).__init__()
        self.fc1 = nn.Conv2d(embedding_dim, width, 1, padding = 0//2)
        self.bn1 = nn.BatchNorm2d(width)
        self.relu1 = nn.ReLU(inplace = True)
        self.fc2 = nn.Conv2d(width, width, 1)
        self.bn2 = nn.BatchNorm2d(width)
        self.relu2 = nn.ReLU(inplace = True)
        self.fc3 = nn.Conv2d(width, 1 , 1)


class ConvProbabilityNet(nn.Module):
    def __init__(self, embedding_dim=4, width = 5, depth = 5):
        super().__init__()
        self.layer1 = MyBlock(embedding_dim + depth,width)
        self.layer2 = MyBlock(embedding_dim ,width)
        self.layer3 = MyBlock(embedding_dim + 2 ,width)
        
        self.positional_embedding_params = (nn.Parameter(torch.rand(depth,3)*10) ).to("cuda")

    

    def forward(self, x):
        # x is C*H*W,E+S+1 (H,W is the window of the crop used here, e.g 100x100, not the image)

        positional_embedding = guide_function(self.positional_embedding_params, width = 100)

        one = self.layer1(torch.cat((x,positional_embedding.expand(x.shape[0],-1,-1,-1)),dim=1))
        two = self.layer2(x)

        output = self.layer3(torch.cat((x,one,two),dim=1))

        return output
    

class MedianFilter(nn.Module):
    def __init__(self, kernel_size: Tuple[int, int]):
        from kornia.filters import MedianBlur
        super(MedianFilter, self).__init__()
        self.kernel_size = kernel_size
        self.MedianBlur = MedianBlur(kernel_size)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.MedianBlur(x)


from einops import rearrange


def feature_engineering(x: torch.Tensor, c: torch.Tensor, sigma: torch.Tensor, window_size: int,
                        mesh_grid_flat: torch.Tensor):
    
    E = x.shape[0]
    h, w = x.shape[-2:]
    C = c.shape[0]
    S = sigma.shape[0]

    x = torch.cat([x,sigma])[:,mesh_grid_flat[0],mesh_grid_flat[1]]
    
    x = rearrange(x, '(E) (C H W) -> C (E) H W', E = E + S, C = C, H = 2 * window_size, W = 2 * window_size)
    c_shaped = c.view(-1, E, 1, 1)
    x[:,:E] -= c_shaped
    x = rearrange(x, 'C (E) H W-> (C H W) (E)', E = E + S)
    return x




def feature_engineering_slow(x: torch.Tensor, c: torch.Tensor, sigma: torch.Tensor, window_size: int,
                        mesh_grid_flat: torch.Tensor):
    
    E = x.shape[0]
    h, w = x.shape[-2:]
    C = c.shape[0]
    S = sigma.shape[0]

    x_slices = x[:, mesh_grid_flat[0], mesh_grid_flat[1]].reshape(E, C, 2 * window_size, 2 * window_size).permute(1, 0, 2,
                                                                                                                  3)  # C,E,2*window_size,2*window_size
    sigma_slices = sigma[:, mesh_grid_flat[0], mesh_grid_flat[1]].reshape(S, C, 2 * window_size, 2 * window_size).permute(1,
                                                                                                                          0,
                                                                                                                          2,
                                                                                                                          3)  # C,S,2*window_size,2*window_size
    c_shaped = c.reshape(-1, E, 1, 1)

    diff = x_slices - c_shaped

    x = torch.cat([diff, sigma_slices], dim=1)  # C,E+S+1,H,W

    x = x.flatten(2).permute(0, -1, 1)  # C,H*W,E+S+1
    x = x.reshape((x.shape[0] * x.shape[1]), x.shape[2])  # C*H*W,E+S+1

    return x



def feature_engineering_2(x: torch.Tensor, xxyy: torch.Tensor, c: torch.Tensor, sigma: torch.Tensor, window_size: int,
                        mesh_grid_flat: torch.Tensor):
    
    # EXTRA DIFF
    E = x.shape[0]
    h, w = x.shape[-2:]
    C = c.shape[0]
    S = sigma.shape[0]

    x_slices = x[:, mesh_grid_flat[0], mesh_grid_flat[1]].reshape(E, C, 2 * window_size, 2 * window_size).permute(1, 0, 2,3)  # C,E,2*window_size,2*window_size
    sigma_slices = sigma[:, mesh_grid_flat[0], mesh_grid_flat[1]].reshape(S, C, 2 * window_size, 2 * window_size).permute(1,
                                                                                                                          0,
                                                                                                                          2,
                                                                                                                          3)  # C,S,2*window_size,2*window_size
    c_shaped = c.reshape(-1, E, 1, 1)

    norm = torch.sqrt(torch.sum(torch.pow(x_slices - c_shaped, 2) + 1e-6, dim=1, keepdim=True))  # C,1,H,W

    diff = x_slices - c_shaped


    x = torch.cat([diff, sigma_slices, norm], dim=1)  # C,E+S+1,H,W

    x = x.flatten(2).permute(0, -1, 1)  # C,H*W,E+S+1
    x = x.reshape((x.shape[0] * x.shape[1]), x.shape[2])  # C*H*W,E+S+1

    return x

def feature_engineering_3(x: torch.Tensor, xxyy: torch.Tensor, c: torch.Tensor, sigma: torch.Tensor, window_size: int,
                        mesh_grid_flat: torch.Tensor):
    
    # NO SIGMA
    E = x.shape[0]
    h, w = x.shape[-2:]
    C = c.shape[0]
    S = sigma.shape[0]

    x_slices = x[:, mesh_grid_flat[0], mesh_grid_flat[1]].reshape(E, C, 2 * window_size, 2 * window_size).permute(1, 0, 2,
                                                                                                                  3)  # C,E,2*window_size,2*window_size
    sigma_slices = sigma[:, mesh_grid_flat[0], mesh_grid_flat[1]].reshape(S, C, 2 * window_size, 2 * window_size).permute(1,
                                                                                                                          0,
                                                                                                                          2,
                                                                                                                          3)  # C,S,2*window_size,2*window_size
    c_shaped = c.reshape(-1, E, 1, 1)

    diff = x_slices - c_shaped


    x = torch.cat([diff, sigma_slices * 0], dim=1)  # C,E+S+1,H,W

    x = x.flatten(2).permute(0, -1, 1)  # C,H*W,E+S+1
    x = x.reshape((x.shape[0] * x.shape[1]), x.shape[2])  # C*H*W,E+S+1

    return x




def feature_engineering_10(x: torch.Tensor, xxyy: torch.Tensor, c: torch.Tensor, sigma: torch.Tensor, window_size: int,
                        mesh_grid_flat: torch.Tensor):
    
    # CONV
    E = x.shape[0]
    h, w = x.shape[-2:]
    C = c.shape[0]
    S = sigma.shape[0]

    x_slices = x[:, mesh_grid_flat[0], mesh_grid_flat[1]].reshape(E, C, 2 * window_size, 2 * window_size).permute(1, 0, 2,
                                                                                                                  3)  # C,E,2*window_size,2*window_size
    sigma_slices = sigma[:, mesh_grid_flat[0], mesh_grid_flat[1]].reshape(S, C, 2 * window_size, 2 * window_size).permute(1,
                                                                                                                          0,
                                                                                                                          2,
                                                                                                                          3)  # C,S,2*window_size,2*window_size
    c_shaped = c.reshape(-1, E, 1, 1)
    diff = x_slices - c_shaped
    x = torch.cat([diff, sigma_slices], dim=1)  # C,E+S+1,H,W

    return x




def feature_engineering_generator(feature_engineering_function):

    if feature_engineering_function == "0" or feature_engineering_function == "7":
        return feature_engineering, 2
    elif feature_engineering_function == "2":
        return feature_engineering_2, 3
    elif feature_engineering_function == "3":
        return feature_engineering_3, 2
    elif feature_engineering_function == "10":
        return feature_engineering_10, 2

    else:
        raise NotImplementedError("Feature engineering function",feature_engineering_function,"is not implemented")





class InstanSeg(nn.Module):

    def __init__(self,
                 n_sigma: int = 1, 
                 instance_weight: float = 1.5, 
                 device: str = 'cuda', 
                 binary_loss_fn_str: str = "lovasz_hinge", 
                 seed_loss_fn = "binary_xloss", 
                 cells_and_nuclei: bool = False, 
                 to_centre: bool = True, 
                 multi_centre: bool = False,
                 window_size = 256, 
                 feature_engineering_function = "0",
                 dim_coords = 2):
        
        super().__init__()
        self.n_sigma = n_sigma
        self.instance_weight = instance_weight
        self.device = device
        self.dim_coords = dim_coords

        self.dim_out = self.dim_coords + self.n_sigma + 1
        self.parameters_have_been_updated = False

        if cells_and_nuclei:
            self.dim_out = self.dim_out * 2
        self.cells_and_nuclei = cells_and_nuclei

        self.to_centre = to_centre
        self.multi_centre = multi_centre
        self.window_size = window_size

        self.num_instance_cap = 50
        self.sort_by_eccentricity = False

        xxyy = generate_coordinate_map(mode = "linear", spatial_dim = self.dim_coords, height = 256, width = 256, device = device)

        self.feature_engineering, self.feature_engineering_width = feature_engineering_generator(feature_engineering_function)
        self.feature_engineering_function = feature_engineering_function


        self.register_buffer("xxyy", xxyy)

        self.update_binary_loss(binary_loss_fn_str)

        self.update_seed_loss(seed_loss_fn)

    def update_binary_loss(self,binary_loss_fn_str):

        if binary_loss_fn_str == "lovasz_hinge":
            from instanseg.utils.loss.lovasz_losses import lovasz_hinge
            def binary_loss_fn(pred, gt, **kwargs):
               # pred = torch.sigmoid_(pred)
                return lovasz_hinge((pred.squeeze(1)), gt,per_image = True)

        elif binary_loss_fn_str == "binary_xloss":
            from instanseg.utils.loss.lovasz_losses import binary_xloss
            self.binary_loss_fn = torch.nn.BCEWithLogitsLoss()
        elif binary_loss_fn_str == "dicefocal_loss":
            from monai.losses import DiceFocalLoss
            
            binary_loss_fn_ = DiceFocalLoss(sigmoid=True)
            def binary_loss_fn(pred, gt, **kwargs):
                l = binary_loss_fn_(pred[None,:,0], gt.unsqueeze(0)) * 1.5
                return l
        elif binary_loss_fn_str == "dice_loss":
            from monai.losses import DiceLoss
            
            binary_loss_fn_ = DiceLoss(sigmoid=True)
            def binary_loss_fn(pred, gt, **kwargs):
                l = binary_loss_fn_(pred[None,:,0], gt.unsqueeze(0)) * 1.5
                return l


        elif binary_loss_fn_str == "general_dice_loss":
            from monai.losses import GeneralizedDiceLoss
            def binary_loss_fn(pred, gt):
                return GeneralizedDiceLoss(sigmoid=True)(pred, gt.unsqueeze(1))
            

        elif binary_loss_fn_str == "cross_entropy":
            from torch.nn import NLLLoss
            assert self.window_size == 256, "Cross entropy loss only works with window size 256"
            assert self.num_instance_cap is None, "Cross entropy loss only works with num_instance_cap = None"
            
            self.l_fn = NLLLoss()
            self.m = nn.LogSoftmax(dim=1)
            
            def binary_loss_fn(pred, gt, sigma):
                pred = torch.cat([sigma[None,None],pred])
                

                gt = torch.cat(((gt.sum(0)==0)[None],gt))
                target = gt.argmax(0)[None]

                pred = pred.squeeze(1).unsqueeze(0)
              
                pred = self.m(pred)

                return self.l_fn(pred,target.long()) * 7
            
        else:
            raise NotImplementedError("Binary loss function",binary_loss_fn,"is not implemented")
        self.binary_loss_fn = binary_loss_fn

    def update_seed_loss(self,seed_loss_fn):
        if seed_loss_fn in ["binary_xloss"]:
            binary_loss = torch.nn.BCEWithLogitsLoss(reduction='none')

            def seed_loss(x,y, mask = None):
                if mask is not None:
                    mask = mask.float()  # Ensure the mask is float for multiplication
                    loss = binary_loss(x, (y > 0).float())  # Calculate the element-wise binary loss
                    masked_loss = loss * mask  # Apply the mask to the loss
                    return masked_loss.sum() / mask.sum()
                else:
                    return binary_loss(x, (y > 0).float()).mean()
                
            self.seed_loss = seed_loss

        elif seed_loss_fn in ["l1_distance"]:
            from instanseg.utils.pytorch_utils import instance_wise_edt
            distance_loss = torch.nn.L1Loss(reduction='none')
            def seed_loss(x,y, mask = None):
                edt = (instance_wise_edt(y.float(), edt_type= 'edt') - 0.5 ) * 15 #This is to mimick the range of CELoss
                loss = distance_loss((x), (edt[None]))

                if mask is not None:
                    mask = mask.float()
                    masked_loss = loss * mask
                    return masked_loss.sum() / mask.sum()
                else:
                    return loss.mean()
                
            self.seed_loss = seed_loss
        else:
            raise NotImplementedError("Seedloss function",seed_loss_fn,"is not implemented")

    def initialize_pixel_classifier(self, model, MLP_width = 10, MLP_input_dim = None):

        if has_pixel_classifier_model(model):
            try:
                self.pixel_classifier = model.pixel_classifier
            except:
                self.pixel_classifier = model.model.pixel_classifier  # This happens when there is an adaptornet
            return model
        else:
            if MLP_input_dim is None:
                MLP_input_dim = self.feature_engineering_width + self.n_sigma -2 + self.dim_coords
            model.pixel_classifier = ProbabilityNet( MLP_input_dim, width = MLP_width)
            if self.feature_engineering_function != "10":
                model.pixel_classifier = ProbabilityNet( MLP_input_dim, width = MLP_width)
            else:
                model.pixel_classifier = ConvProbabilityNet( MLP_input_dim, width = MLP_width)
            self.pixel_classifier = model.pixel_classifier.to(self.device)
            return model

    def forward(self, prediction: torch.Tensor, instances: torch.Tensor, w_inst: float = 1.5, w_seed: float = 1.0):

        w_inst = self.instance_weight

        batch_size, height, width = prediction.size(
            0), prediction.size(2), prediction.size(3)

        xxyy = self.xxyy[:, 0:height, 0:width].contiguous()  # 2 x h x w

        loss = 0
  
        if self.cells_and_nuclei:
            dim_out = int(self.dim_out / 2)
        else:
            dim_out = self.dim_out

        for mask_channel in range(0, instances.shape[1]):

            if mask_channel == 0:
                prediction_b = prediction[:, 0: dim_out, :, :]
            else:
                prediction_b = prediction[:, dim_out:, :, :]

            instances_batch = instances

            if not self.to_centre:

                spatial_emb_batch = (torch.sigmoid((prediction_b[:, 0: self.dim_coords]))-0.5) * 8 + xxyy
            else:
                spatial_emb_batch = (prediction_b[:, 0: self.dim_coords]) + xxyy
            sigma_batch = prediction_b[:, self.dim_coords: self.dim_coords + self.n_sigma]  # n_sigma x h x w
            seed_map_batch = prediction_b[:, self.dim_coords + self.n_sigma: self.dim_coords + self.n_sigma + 1]  # 1 x h x w

            for b in range(0, batch_size):

                spatial_emb = spatial_emb_batch[b]
                sigma = sigma_batch[b]
                seed_map = seed_map_batch[b]

                instance_loss = 0
                seed_loss = 0

                instance = instances_batch[b, mask_channel].unsqueeze(0)  # 1 x h x w

                if (instance < 0).all(): #-1 means not annotated
                    continue
       

                elif instance.min() < 0: #label is sparse
                    mask = instance >=0
                    instance[instance < 0] = 0
                else:
                    mask = None

                seed_loss_tmp = self.seed_loss(seed_map,instance, mask = mask)

                seed_loss += seed_loss_tmp

                if w_inst == 0:
                    loss += w_seed * seed_loss
                    continue

                instance_ids = instance.unique()
                instance_ids = instance_ids[instance_ids != 0]

                if len(instance_ids) > 0:

                    instance = torch_fastremap(instance)

                    onehot_labels = torch_onehot(instance).squeeze(0)  # C x h x w

                    if self.num_instance_cap is not None: #This is to cap the number of objects to avoid OOM errors.
                         if self.num_instance_cap < onehot_labels.shape[0]:
                            if self.sort_by_eccentricity:
                                eccentricities = eccentricity_batch(onehot_labels.float())
                                idx = eccentricities.argsort(descending = True)[:self.num_instance_cap]
                            else:
                                idx = torch.randperm(onehot_labels.shape[0])[:self.num_instance_cap]
                            onehot_labels = onehot_labels[idx]


                    if self.multi_centre:
                        seed_map_tmp = torch.sigmoid(seed_map)
                        
                        centroids = torch_peak_local_max(seed_map_tmp.squeeze() * onehot_labels.sum(0), neighbourhood_size = 3, minimum_value = 0.5).T

                        if self.to_centre:
                            centres = xxyy[:,centroids[0],centroids[1]].detach().T
                        else:
                            centres = spatial_emb[:,centroids[0],centroids[1]].detach().T

                        idx = torch.randperm(centroids.shape[1])[:self.num_instance_cap]
                  
                        centres = centres[idx]
                        centroids = centroids[:,idx]

                        instance_labels = onehot_labels[:,centroids[0],centroids[1]].float().argmax(0)
                        onehot_labels = onehot_labels[instance_labels]

                        centroids = centroids.T
             
                    else:
                        if self.to_centre:

                            seed_map_min = seed_map.min()
           
                            seed_map_tmp = (seed_map - seed_map.min()).detach()
                            centres = xxyy.flatten(1).T[((seed_map_tmp * onehot_labels).flatten(1)).argmax(1)]  # location at max seed (used in postprocessing)
                            seed_map = seed_map_tmp + seed_map_min
                        else:
                            seed_map_tmp = seed_map - seed_map.min()
                            centres = spatial_emb.flatten(1).T[((seed_map_tmp * onehot_labels).flatten(1)).argmax(1)].detach()  # embedding at max seed (used in postprocessing)
                            
                        centroids = (torch.sum(((xxyy[:2] * onehot_labels.unsqueeze(1))).flatten(2),dim=2)/onehot_labels.flatten(1).sum(1)[:,None] )* (256 / 64) #coordinates of centre of mass
                        centroids = torch.stack((centroids[:, 1], centroids[:, 0])).T

                    if len(centroids) == 0:
                        loss += w_seed * seed_loss
                        continue
      
                    dist, coords = compute_crops(spatial_emb, 
                                                 centres, 
                                                 sigma, 
                                                 centroids, 
                                                 feature_engineering = self.feature_engineering,
                                                 pixel_classifier=self.pixel_classifier,
                                                 window_size = self.window_size)
                    

                    crop = onehot_labels.squeeze(1)[coords[0], coords[1], coords[2]].reshape(-1, self.window_size, self.window_size)

                    instance_loss = instance_loss + self.binary_loss_fn(dist,crop.float(), sigma = sigma[0])

                loss += w_inst * instance_loss + w_seed * seed_loss


        loss = loss / (b + 1)

        if self.cells_and_nuclei:
            loss = loss / 2

        return loss
    
    

    def update_hyperparameters(self,params):
        self.parameters_have_been_updated = True
        self.params = params



    #@timer
    def postprocessing(self, prediction: Union[torch.Tensor, np.ndarray],
                        mask_threshold: float = 0.53,
                        peak_distance: int = 5,
                        seed_threshold: float = 0.8,
                        overlap_threshold: float = 0.3,
                        mean_threshold: float = 0.1,
                        window_size: int = 128,
                        min_size = 10,
                       device=None,
                       classifier=None,
                       cleanup_fragments: bool = False,
                       max_seeds: int = 2000,
                       return_intermediate_objects: bool = False,
                       precomputed_crops: torch.Tensor = None,
                       precomputed_seeds: torch.Tensor = None,
                       img=None):

        if device is None:
            device = self.device
        if classifier is None:
            classifier = self.pixel_classifier

        if self.parameters_have_been_updated:
            mask_threshold = self.params['mask_threshold']
            peak_distance = self.params['peak_distance']
            seed_threshold = self.params['seed_threshold']
            overlap_threshold = self.params['overlap_threshold']
            if "min_size" in self.params:
                min_size = self.params['min_size']
            if "mean_threshold" in self.params:
                mean_threshold = self.params['mean_threshold']
            

        if isinstance(prediction, np.ndarray):
            prediction = torch.tensor(prediction, device=device)

        if self.cells_and_nuclei:
            iterations = 2
            dim_out = int(self.dim_out / 2)
        else:
            iterations = 1
            dim_out = self.dim_out

        labels = []

        for i in range(iterations):

            if precomputed_crops is None:

                if i == 0:
                    prediction_i = prediction[0: dim_out, :, :]
                else:
                    prediction_i = prediction[dim_out:, :, :]

                height, width = prediction_i.size(1), prediction_i.size(2)

                ##torch.cuda.synchronize()

                xxyy = generate_coordinate_map(mode = "linear", spatial_dim = self.dim_coords, height = height, width = width, device = device)

                #torch.cuda.synchronize()

                if not self.to_centre:
                    fields = (torch.sigmoid(prediction_i[0:self.dim_coords])-0.5) * 8
                else:
                    fields = prediction_i[0:self.dim_coords]

                sigma = prediction_i[self.dim_coords:self.dim_coords + self.n_sigma]
                mask_map = torch.sigmoid(prediction_i[self.dim_coords + self.n_sigma])

                if (mask_map > mask_threshold).max() == 0:  # no foreground pixels
                    label = torch.zeros(mask_map.shape, dtype=int, device=mask_map.device).squeeze()
                    labels.append(label)
                    continue

            # local_centroids_idx = #torch.tensor([[20,21,32,32,34],[30,35,36,364,346]],device = device).long().T

                #torch.cuda.synchronize()

                if precomputed_seeds is None:
                    local_centroids_idx = torch_peak_local_max(mask_map, neighbourhood_size=int(peak_distance), minimum_value=seed_threshold)
                else:
                    local_centroids_idx = precomputed_seeds


                #torch.cuda.synchronize()

                fields = fields + xxyy
                if self.to_centre:
                    fields_at_centroids = xxyy[:, local_centroids_idx[:, 0], local_centroids_idx[:, 1]]
                else:
                    fields_at_centroids = fields[:, local_centroids_idx[:, 0], local_centroids_idx[:, 1]]

                if local_centroids_idx.shape[0] > max_seeds:
                    print("Too many seeds, skipping", local_centroids_idx.shape[0])
                    label = torch.zeros(mask_map.shape, dtype=int, device=mask_map.device).squeeze()
                    labels.append(label)
                    continue

                
                C = fields_at_centroids.shape[0]

                h, w = mask_map.shape[-2:]
                window_size = min(window_size, h, w)
                window_size = window_size - window_size % 2

                if C == 0:
                    label = torch.zeros(mask_map.shape, dtype=int, device=mask_map.device).squeeze()
                    labels.append(label)
                    continue

                #torch.cuda.synchronize()
                crops, coords = compute_crops(fields, 
                                                fields_at_centroids.T, 
                                                sigma, 
                                                local_centroids_idx.int(), 
                                                feature_engineering = self.feature_engineering,
                                                pixel_classifier=self.pixel_classifier,
                                                window_size=window_size) # about 65% of the time
                #torch.cuda.synchronize()
                coords = coords[1:] # The first channel are just channel indices, not required here.

                if return_intermediate_objects:
                    return crops, coords, mask_map

                C = crops.shape[0]
                if C == 0:
                    label = torch.zeros(mask_map.shape, dtype=int, device=mask_map.device).squeeze()
                    labels.append(label)
                    continue

                

            else:
                crops,coords,mask_map = precomputed_crops
                C = crops.shape[0]



            h, w = mask_map.shape[-2:]

            label = merge_sparse_predictions(crops, coords, mask_map, size=(C,h, w), mask_threshold=mask_threshold, window_size=window_size, min_size=min_size, overlap_threshold=overlap_threshold, mean_threshold=mean_threshold).int() #about 30% of the time


            # from utils.pytorch_utils import centroids_from_lab
            # centroids, ids = centroids_from_lab(label)
            # coords = centre_crop(centroids=centroids, window_size=window_size, h=h, w=w)
            # crops = label[...,coords[0],coords[1]].view(centroids.shape[0],1,window_size,window_size)
            # clean = (crops == ids[1:][:,None,None,None]) * ids[1:][:,None,None,None]

            # clean = connected_components((clean > 0).float(),num_iterations= 128)

            # for ii, cc_map in enumerate(clean):
            #     l= torch.unique(cc_map[cc_map>0], sorted=True)
            #     clean[ii] = cc_map == l[-1]
            # clean = clean.int()
            # clean = clean * torch.arange(clean.shape[0],device = clean.device,dtype = label.dtype)[:,None,None,None]

            # label = convert(clean,coords,size = [h,w]).int()



            labels.append(label.squeeze())


        if len(labels) == 1:
            return labels[0][None]  # 1,H,W
        else:
            return torch.stack(labels)  # 2,H,W
        


    def TTA_postprocessing(self, img, model, transforms,
                        mask_threshold: float = 0.53,
                        peak_distance: int = 5,
                        seed_threshold: float = 0.8,
                        overlap_threshold: float = 0.3,
                        mean_threshold: float = 0.1,
                        window_size: int =64,
                        min_size = 10,
                       device=None,
                       classifier=None,
                       cleanup_fragments: bool = False,
                       reduction = "mean",
                       max_seeds: int = 2000,):

        

        cells_and_nuclei = self.cells_and_nuclei
        if self.cells_and_nuclei:
            iterations = 2
            assert self.dim_out % 2 == 0,  print("The model should an even number of output channels for cells and nuclei.")
            dim_out = int(self.dim_out / 2)
        else:
            iterations = 1
            dim_out = self.dim_out

        out_labels = []

        transforms = [t for t in transforms] + [IdentityTransform()]
        
        for i in range(iterations):

            all_masks_list = []
            all_predictions = []

            self.cells_and_nuclei = False

            
            for t in transforms:
                with torch.cuda.amp.autocast():
                    augmented_image = t.augment_image(img)
                    augmented_image, pad = instanseg_padding(augmented_image, extra_pad= 0, min_dim = 32)
                    prediction = model(augmented_image)[:,i * dim_out:(i+1) * dim_out]
                    prediction = recover_padding(prediction, pad)
                    mask_map = prediction[:,-1][None] 
                    mask_map = t.deaugment_mask(mask_map)
                #  show_images(mask_map)
                    all_masks_list.append(mask_map.cpu())
                    all_predictions.append(prediction.cpu())

          #  pdb.set_trace()

            if reduction == "local_max":

                local_maxima_maps = [torch_peak_local_max(mask.squeeze().float().to(device), int(peak_distance),seed_threshold, return_map = True) for mask in all_masks_list]
                local_maxima_map = torch_peak_local_max(torch.stack(local_maxima_maps).max(0)[0].squeeze(),int(peak_distance),seed_threshold, return_map = True)
                all_masks = torch.mean(torch.stack(all_masks_list),dim=0).float().to(device)

            elif reduction in ["mean", "median"]:
                if reduction == "mean":
                    all_masks = torch.mean(torch.stack(all_masks_list),dim=0).float().to(device)
                elif reduction == "median":
                    all_masks = torch.median(torch.stack(all_masks_list),dim=0)[0].to(device)

                local_maxima_map = torch_peak_local_max(all_masks.squeeze(), neighbourhood_size=int(peak_distance), minimum_value=seed_threshold, return_map = True)
            
            local_maxima_map = (local_maxima_map > 0).float()
            centroids = torch.stack(torch.where(local_maxima_map.squeeze())).T


            if len(centroids) == 0:
                out_labels.append(torch.zeros((1,*all_masks.shape[-2:]), dtype=int, device=device))
                continue

            local_maxima_map[...,centroids[:,0],centroids[:,1]] = torch.arange(1,centroids.shape[0]+1,device = local_maxima_map.device).float()


            all_crops = []

            for (t, prediction) in (zip(transforms, all_predictions)):
                prediction_tmp = prediction.clone().float().to(device)
                prediction_tmp[:,-1] = t.augment_image(all_masks)
                prediction_tmp = prediction_tmp.squeeze(0)

                local_maxima_map_tmp = t.augment_image( local_maxima_map )
                centroids = torch.stack(torch.where(local_maxima_map_tmp.squeeze())).T
                values = local_maxima_map_tmp[...,centroids[:,0],centroids[:,1]]
                centroids = centroids[values.sort()[1]][0,0]

                out = self.postprocessing(prediction_tmp, mask_threshold, peak_distance, seed_threshold, overlap_threshold, mean_threshold, window_size, min_size, device, classifier, 
                                                            cleanup_fragments, max_seeds, return_intermediate_objects = True, precomputed_seeds = centroids)
                
                if len(out)==3:
                    crops, coords, mask_map = out
                else:
                    pdb.set_trace()

                crops = t.deaugment_mask(crops)
                all_crops.append(crops.cpu())


         #   show_images(torch.cat([*torch.cat(all_crops,dim = 3)[:50]],dim = 1),colorbar= False)


            all_crops = torch.median(torch.stack(all_crops).float(),dim=0)[0].to(device)
         #   all_crops = torch.mean(torch.stack(all_crops).float(),dim=0).to(device)
         #   all_crops = torch.max(torch.stack(all_crops).float(),dim=0)[0].to(device)


            labels = self.postprocessing(prediction, mask_threshold, peak_distance, seed_threshold, overlap_threshold, mean_threshold, window_size, min_size, device, classifier,
                                        cleanup_fragments, max_seeds, precomputed_crops = (all_crops, coords, mask_map))

            
            out_labels.append(labels)
        self.cells_and_nuclei = cells_and_nuclei

        labels = torch.stack(out_labels, dim = 1).squeeze(0)
        #show_images(labels)
        return labels


        
class IdentityTransform:
    def augment_image(self, img):
        return img
    def deaugment_mask(self, mask):
        return mask

        

from instanseg.utils.biological_utils import resolve_cell_and_nucleus_boundaries

class InstanSeg_Torchscript(nn.Module):
    def __init__(self, model, 
                 cells_and_nuclei: bool = False,
                 pixel_size : float = 0, 
                 n_sigma: int = 2, 
                 dim_coords:int = 2, 
                 to_centre: bool = True,
                 backbone_dim_in: int = 3,  
                 feature_engineering_function:str  = "0",
                 params = None,
                 mixed_precision: bool = False):
        super(InstanSeg_Torchscript, self).__init__()

        model.eval()

        with torch.cuda.amp.autocast(enabled=mixed_precision):
            with torch.no_grad():
                self.fcn = torch.jit.trace(model, torch.rand(1, backbone_dim_in, 256, 256))

        try:
            self.pixel_classifier = model.pixel_classifier
        except:
            self.pixel_classifier = model.model.pixel_classifier  # I think this is a pytorch version issue between 1.13.1 and 2.0.0
        self.cells_and_nuclei = cells_and_nuclei
        self.pixel_size = pixel_size
        self.dim_coords = dim_coords
        self.n_sigma = n_sigma
        self.to_centre = to_centre
        self.feature_engineering, self.feature_engineering_width = feature_engineering_generator(feature_engineering_function)
        self.params = params,
    
        self.index_dtype = torch.long #torch.int

        # self.traced_feature_engineering = torch.jit.trace(self.feature_engineering, 
        #                                                   (torch.ones(self.dim_coords,256,256).float(), 
        #                                                    torch.rand(55, 2).int(), 
        #                                                    torch.ones(self.n_sigma,256,256).float(), 
        #                                                    torch.tensor([32]).int(), 
        #                                                    (torch.rand(2, 225280) * 256).int()))

        #median_filter = MedianFilter((3,3))

        # if self.params is not None:
        #     for key in self.params:
        #         setattr(self, key, self.params[key])

    def forward(self, x: torch.Tensor,
                target_segmentation : torch.Tensor = torch.tensor([1,1]), # Nuclei / Cells
                min_size:int = 10,
                mask_threshold: float = 0.53,
                peak_distance: int = 5,
                seed_threshold: float = 0.7,
                overlap_threshold: float = 0.3,
                mean_threshold: float  = 0.0,
                window_size: int = 32,
                cleanup_fragments: bool= False,
                resolve_cell_and_nucleus: bool = True,
                ) -> torch.Tensor:
        

        torch.clamp_max_(x, 3) #Safety check, please normalize inputs properly!
        torch.clamp_min_(x, -2)

        x, pad = instanseg_padding(x, extra_pad=0)


       # with torch.autocast(device_type='cuda', dtype=torch.float16):
        with torch.no_grad():
            x_full = self.fcn(x)
    
            dim_out = x_full.shape[1]

            if self.cells_and_nuclei:
                iterations = torch.tensor([0,1]) [target_segmentation.squeeze().to("cpu") > 0 ]
                dim_out = int(dim_out / 2)

            else:
                iterations = torch.tensor([0])
                dim_out = dim_out

            output_labels_list = []

            for image_index in range(x_full.shape[0]):

            
                labels_list = []

                for i in iterations:

                    if i == 0:

                        x = x_full[image_index,0: dim_out, :, :]
                    else:
                        x = x_full[image_index,dim_out:, :, :]

                    x = recover_padding(x, pad)

                    height, width = x.size(1), x.size(2)

                    xxyy = generate_coordinate_map(mode = "linear", spatial_dim = self.dim_coords, height = height, width = width, device = x.device)


                    if not self.to_centre:
                        fields = (torch.sigmoid(x[0:self.dim_coords])-0.5) * 8
                    else:
                        fields = x[0:self.dim_coords] 

                
                    sigma = x[self.dim_coords:self.dim_coords + self.n_sigma]
                    mask_map = torch.sigmoid(x[self.dim_coords + self.n_sigma])

                    centroids_idx = torch_peak_local_max(mask_map, neighbourhood_size=peak_distance,
                                                        minimum_value=seed_threshold, dtype= self.index_dtype)  # .to(prediction.device)
                    #num_initial_centroids = centroids_idx.shape[0]


                    fields = fields + xxyy

                    if self.to_centre:
                        fields_at_centroids = xxyy[:, centroids_idx[:, 0], centroids_idx[:, 1]]
                    else:
                        fields_at_centroids = fields[:, centroids_idx[:, 0], centroids_idx[:, 1]]


                    x = fields
                    c = fields_at_centroids.T
                    E = x.shape[0]
                    h, w = x.shape[-2:]
                    C = c.shape[0]
                    S = sigma.shape[0]

                    if C == 0:
                        label = torch.zeros(mask_map.shape, dtype= torch.float32, device=mask_map.device).squeeze()
                        labels_list.append(label)
                        continue

                    window_size = window_size
                    centroids = centroids_idx.clone().cpu()  # C,2
                    centroids[:, 0].clamp_(min=window_size, max=h - window_size)
                    centroids[:, 1].clamp_(min=window_size, max=w - window_size)
                    window_slices = centroids[:, None].to(x.device) + torch.tensor([[-1, -1], [1, 1]] , device = x.device, dtype=centroids.dtype) * window_size
                    window_slices = window_slices  # C,2,2

                    slice_size = window_size * 2

                    # Create grids of indices for slice windows
                    grid_x, grid_y = torch.meshgrid(
                        torch.arange(slice_size, device=x.device, dtype=self.index_dtype),
                        torch.arange(slice_size, device=x.device, dtype=self.index_dtype), indexing="ij")
                    mesh = torch.stack((grid_x, grid_y))

                    mesh_grid = mesh.expand(C, 2, slice_size, slice_size)  # C,2,2*window_size,2*window_size
                    mesh_grid_flat = torch.flatten(mesh_grid, 2).permute(1, 0, -1)  # 2,C,2*window_size*2*window_size
                    idx = window_slices[:, 0].permute(1, 0)[:, :, None]
                    mesh_grid_flat = mesh_grid_flat + idx
                    mesh_grid_flat = torch.flatten(mesh_grid_flat, 1)  # 2,C*2*window_size*2*window_size

                #    x = self.traced_feature_engineering(x, c, sigma, torch.tensor(window_size).int(), mesh_grid_flat)
                    x = feature_engineering_slow(x, c, sigma, torch.tensor(window_size).int(), mesh_grid_flat)

                    x = torch.sigmoid(self.pixel_classifier(x))

                    x = x.reshape(C, 1, slice_size, slice_size)

                
                    C = x.shape[0]

                    if C == 0:
                        label = torch.zeros(mask_map.shape, dtype= torch.float32, device=mask_map.device).squeeze()
                        labels_list.append(label)
                        continue

                    original_device = x.device

                    if x.is_mps:
                        device = 'cpu'
                        mesh_grid_flat = mesh_grid_flat.to(device)
                        x = x.to(device)
                        mask_map = mask_map.to(device)

                    coords = mesh_grid_flat.reshape(2, C, slice_size, slice_size)

                    if cleanup_fragments:

                        top_left = window_slices[:,0,:]
                        shifted_centroid = centroids_idx - top_left
                        cc = connected_components((x>mask_threshold).float(),num_iterations= 64)
                        labels_to_keep = cc[torch.arange(cc.shape[0]),0,shifted_centroid[:,0],shifted_centroid[:,1]]
                        in_mask = cc == labels_to_keep[:,None,None,None]
                        x *= in_mask

                

                    labels = convert(x, coords, size=(h, w), mask_threshold=mask_threshold)[None]

            

                    idx = torch.arange(1, C + 1, device=x.device, dtype = self.index_dtype)
                    stack_ID = torch.ones((C, slice_size, slice_size), device=x.device, dtype=self.index_dtype)
                    stack_ID = stack_ID * (idx[:, None, None] - 1)

                    iidd = torch.stack((stack_ID.flatten(), mesh_grid_flat[0] * w + mesh_grid_flat[1]))

                    fg = x.flatten() > mask_threshold
                    x = x.flatten()[fg]
                    sparse_onehot = torch.sparse_coo_tensor(
                        iidd[:, fg],
                        (x.flatten() > mask_threshold).float(),
                        size=(C, h * w),
                        dtype=x.dtype,
                        device=x.device
                    )

                    object_areas = torch.sparse.sum(sparse_onehot.to(torch.bool).float(), dim=(1,)).values()
                    sum_mask_value = torch.sparse.sum((sparse_onehot * mask_map.flatten()[None]), dim=(1,)).values()
                    mean_mask_value = sum_mask_value / object_areas
                    objects_to_remove = ~torch.logical_and(mean_mask_value > mean_threshold, object_areas > min_size)

                    iou = fast_sparse_iou(sparse_onehot)

                    remapping = find_connected_components((iou > overlap_threshold).to(self.index_dtype))
                    
                    labels = remap_values(remapping, labels)


                    labels_to_remove = (torch.arange(0, len(objects_to_remove), device=objects_to_remove.device) + 1)[
                        objects_to_remove]
                    labels[torch.isin(labels, labels_to_remove)] = 0

                    


                    labels_list.append(labels.squeeze().to(original_device))

                if len(labels_list) == 1:
                    lab = labels_list[0][None, None]  # 1,1,H,W
                else:
                    lab = torch.stack(labels_list)[None] 

                if lab.shape[1] == 2 and resolve_cell_and_nucleus: #nuclei and cells
                    lab = resolve_cell_and_nucleus_boundaries(lab)

                output_labels_list.append(lab[0])
            
            lab = torch.stack(output_labels_list) # B,C,H,W

            


            return lab.to(torch.float32) # B,C,H,W

if __name__ == "__main__":

    import sys

    import torch
    from instanseg.utils.create_bioimageio_model import export_bioimageio
    from instanseg.utils.utils import show_images

    from instanseg.utils.utils import export_to_torchscript

    model_name = "1804630"

    export_to_torchscript(model_name, show_example=True, mixed_predicision=False, model_path = "../../models/")


    pdb.set_trace()









from __future__ import annotations


import torch
import torch.nn as nn

def create_gaussian_grid(N, sigma, device = "cuda",channels = 1):
    # Create a grid of coordinates centered at (0, 0)
    x = torch.linspace(-N//2, N//2, N, device = device)
    y = torch.linspace(-N//2, N//2, N, device = device)
    xx, yy = torch.meshgrid(x, y, indexing="ij")
    
    # Compute the Gaussian values
    gaussian_values = torch.exp(-(xx**2 + yy**2) / (2 * sigma**2)).repeat(channels,1,1)
    
    return gaussian_values


from torch import nn
from einops import rearrange
import torch.nn.functional as F


class LocalInstanceNorm(nn.Module):
    def __init__(self,in_channels = 1):
        super(LocalInstanceNorm, self).__init__()
        self.kernel_size = int(64/(in_channels//32))
        self.norm = nn.InstanceNorm1d(in_channels, affine=True)
        self.sigma =  int(64/(in_channels//32))#30#torch.nn.Parameter(torch.tensor(float(sigma),device = "cuda"),requires_grad = True)
        self.gaussian = create_gaussian_grid(self.kernel_size, sigma= self.sigma,device ="cuda",channels = in_channels).flatten().view(1,-1,1) + 1e-5

    def forward(self, x):
        assert x.dim() == 4, print("Only implemented for batch 4d tensor",x.dim())

        b,c,h,w = x.shape
        x = F.unfold(x, kernel_size=(self.kernel_size,self.kernel_size), stride=self.kernel_size//2)
        x = rearrange(x, 'b (c k1 k2) n -> (b n) c (k1 k2)', c = c, k1 = self.kernel_size, k2 = self.kernel_size)
        #x = self.norm(x)

        x = x - x.mean(dim = 1,keepdim = True)
        x = x / (x.std(dim = 1,keepdim = True) + 1e-5)

        x = rearrange(x, '(b n) c (k1 k2) -> b (c k1 k2) n', b= b, c = c, k1 = self.kernel_size, k2 = self.kernel_size)

        x = x * self.gaussian
        counter = F.fold(torch.ones_like(x) * self.gaussian, (h,w),kernel_size=(self.kernel_size,self.kernel_size), stride=self.kernel_size//2)#.view(1,1,10,10,-1)
        x = F.fold(x, (h,w), kernel_size=(self.kernel_size,self.kernel_size), stride=self.kernel_size//2)
        
        result = x / counter

        return result


def conv_norm_act(in_channels, out_channels, sz,norm, act = "ReLU"):

    if norm == "None" or norm is None:
        norm_layer = nn.Identity()
    elif norm.lower() == "batch":
        norm_layer = nn.BatchNorm2d(out_channels,eps = 1e-5, momentum = 0.05)
    elif norm.lower() == "instance":
        norm_layer = nn.InstanceNorm2d(out_channels,eps = 1e-5, track_running_stats=False, affine = True)
    elif norm.lower() == "local":
        norm_layer = LocalInstanceNorm(in_channels = out_channels)
    else:
        raise ValueError("Norm must be None, batch or instance")
    

    if act == "None" or act is None:
        act_layer = nn.Identity()
    elif act.lower() == "relu":
        act_layer = nn.ReLU(inplace=True)
    elif act.lower() == "mish":
        act_layer = nn.Mish(inplace=True)
    else:
        raise ValueError("Act must be None, ReLU or Mish")
    
    return nn.Sequential(
        nn.Conv2d(in_channels, out_channels, sz, padding=sz//2),
        norm_layer,
        act_layer,
    )  


import torch.nn.functional as F
class DecoderBlock(nn.Module):
    def __init__(
            self,
            in_channels,
            skip_channels,
            out_channels,
            norm = "BATCH",
            act = "ReLU",
            shallow = False,
    ):
        super().__init__()

        
        self.conv0 = conv_norm_act(in_channels, out_channels, 1,norm,act)
        self.conv_skip = conv_norm_act(skip_channels, out_channels, 1,norm,act)
        self.conv1 = conv_norm_act(in_channels, out_channels, 3,norm,act)
        self.conv2 = conv_norm_act(out_channels, out_channels, 3,norm,act)
        self.conv3 = conv_norm_act(out_channels, out_channels, 3,norm,act)
        self.conv4 = conv_norm_act(out_channels, out_channels, 3,norm,act)

        if shallow:
            self.conv3 = nn.Identity()

    def forward(self, x, skip=None):
        x = F.interpolate(x, scale_factor=2, mode="nearest")
        proj = self.conv0(x)
        x = self.conv1(x)
        x =  proj + self.conv2(x + self.conv_skip(skip))
        x = x + self.conv4(self.conv3(x))
        return x


class EncoderBlock(nn.Module):
    def __init__(
            self,
            in_channels,
            out_channels,
            pool = True,
            norm = "BATCH",
            act = "ReLU",
            shallow = False,
    ):
        super().__init__()

        if pool:
            self.maxpool = nn.MaxPool2d(2, 2)
        else:
            self.maxpool = nn.Identity()
        self.conv0 = conv_norm_act(in_channels, out_channels, 1,norm,act)
        self.conv1 = conv_norm_act(in_channels, out_channels, 3,norm,act)
        self.conv2 = conv_norm_act(out_channels, out_channels, 3,norm,act)
        self.conv3 = conv_norm_act(out_channels, out_channels, 3,norm,act)
        self.conv4 = conv_norm_act(out_channels, out_channels, 3,norm,act)

        if shallow:
            self.conv2 = nn.Identity()
            self.conv3 = nn.Identity()

    def forward(self, x):

        x = self.maxpool(x)
        proj = self.conv0(x)
        x = self.conv1(x)
        x =  proj + self.conv2(x)
        x = x + self.conv4(self.conv3(x))
        return x
    
class Decoder(nn.Module):
    def __init__(
            self,
            layers,
            out_channels,
            norm,
            act):
        super().__init__()
        
        self.decoder = nn.ModuleList([DecoderBlock(layers[i],layers[i+1],layers[i+1],norm = norm, act =  act) for i in range(len(layers)-1)])
        
        self.final_block = nn.ModuleList([conv_norm_act(layers[-1],out_channel,1, norm = norm if (norm is not None) and norm.lower() != "instance" else None,act = None) for out_channel in out_channels])
        
    def forward(self,x,skips):
        for layer,skip in zip(self.decoder,skips[::-1]):
            x = layer(x,skip)

        x = torch.cat([final_block(x) for final_block in self.final_block],dim = 1)
        return x
            
    
class InstanSeg_UNet(nn.Module):
    def __init__(self,in_channels,out_channels,layers = [256,128,64,32],norm = "BATCH",dropout = 0, act = "ReLu"):
        super().__init__()
        layers = layers[::-1]
        self.encoder = nn.ModuleList([EncoderBlock(in_channels,layers[0],pool = False,norm = norm, act = act)] + [EncoderBlock(layers[i],layers[i+1],norm = norm, act = act) for i in range(len(layers)-1)])
        layers = layers[::-1]

        # out_channels should be a list of lists [[2,2,1],[2,2,1]] means two decoders, each with 3 output blocks. The output will be of shape 10.

        if type(out_channels) == int:
            out_channels = [[out_channels]]
        if type(out_channels[0]) == int:
            out_channels = [out_channels]
            
        self.decoders = nn.ModuleList([Decoder(layers,out_channel,norm, act) for out_channel in out_channels])
    
    def forward(self,x):
        skips = []
        for n,layer in enumerate(self.encoder):
            x = layer(x)
            if n < len(self.encoder)-1:
                skips.append(x)

        return torch.cat([decoder(x,skips) for decoder in self.decoders],dim = 1)


if __name__ == "__main__":

    import pdb

    net = InstanSeg_UNet(
        in_channels=3,
        out_channels=[5,3],
    )

    print(net(torch.randn(1,3,256,256)).shape)
